__author__ = 'David Randolph'
# Copyright (c) 2014-2018 David A. Randolph.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
from tatsu import parse


class DAnnotation:
    GRAMMAR = """
        @@grammar::Calc

        sequence = upper:staff ['@' lower:staff] ;

        staff = '&'.{line};
        line = {score_fingering}* ;

        score_fingering = orn:ornamental ["/" alt_orn:ornamental]
        | pf:pedaled_fingering ['/' alt_pf:pedaled_fingering]
        | p:pedaling ['/' alt_p:pedaling]
        ;

        ornamental = ornaments:('(' {pedaled_fingering}+ ')') ;

        pedaled_fingering = soft:[soft] fingering:fingering damper:[damper] segmenter:[segmenter];
        pedaling = soft:{soft}+ 'x' damper:{damper}+ ;

        fingering = strike:finger ['-' release:finger] ;
        finger = hand:[hand] digit:digit ;

        segmenter = "," | ";" | "." ;
        damper = '_' | '^' ;
        soft = 'p' | 'f' ;
        hand = '<' | '>' ;
        digit = '1' | '2' | '3' | '4' | '5' | 'x' ;
    """

    @staticmethod
    def ast_for_abcdf(abcdf):
        ast = parse(DAnnotation.GRAMMAR, abcdf)
        return ast

    def parse(self):
        return self._ast

    def score_fingering_at_index(self, index, staff="upper"):
        if staff != "upper" and staff != "lower":
            raise Exception("Score fingerings ordered within upper and lower staff only.")
        if staff == "upper":
            lines = self._ast.upper
        else:
            lines = self._ast.lower

        offset = 0
        for line in lines:
            if index < len(line) + offset:
                adjusted_index = index - offset
                return line[adjusted_index]
            offset += len(line)

    def strike_digit_at_index(self, index, staff="upper"):
        sf = self.score_fingering_at_index(staff=staff, index=index)
        strike = sf.pf.fingering.strike
        return strike.digit

    def parse_upper(self):
        upper_abcdf = self.upper_abcdf()
        return DAnnotation.ast_for_abcdf(upper_abcdf)

    def parse_lower(self):
        lower_abcdf = self.upper_abcdf()
        return DAnnotation.ast_for_abcdf(lower_abcdf)

    def score_fingering_count(self, staff="both"):
        # print(staff + ":::" + self.abcdf() + "\n")
        ast = self.parse()
        # print(ast)
        count = 0
        # Each staff is parsed into an array of lines. Each
        # line is an array of "score fingerings," or note
        # fingerings with all the trimmings.
        if 'upper' in ast and (staff == "upper" or staff == "both"):
            lines = ast['upper']
            for line in lines:
                count += len(line)
        if 'lower' in ast and (staff == "lower" or staff == "both"):
            lines = ast['lower']
            for line in lines:
                count += len(line)
        return count

    def segregated_strike_digits(self, staff="upper", hand=None):
        """
        :return: String of digits (1-5), assuming all fingerings are
                 are for the specified hand (">" or right for the
                 upper staff by default).

                 Returns None if any fingerings for the other hand
                 are detected.
        """
        if staff not in ("upper", "lower"):
            raise Exception("Invalid input: staff must be 'upper' or 'lower'.")

        if not hand:
            hand = ">"
            if staff == "lower":
                hand = "<"

        digits = []
        ast = self.parse()
        if staff == "upper":
            lines = ast.upper
        else:
            lines = ast.lower

        for line in lines:
            for score_fingering in line:
                strike = score_fingering.pf.fingering.strike
                current_hand = strike.hand
                digit = strike.digit
                if current_hand and current_hand != hand:
                    return None
                digits.append(digit)
        digit_str = "".join(digits)
        return digit_str

    def handed_strike_digits(self, staff="upper"):
        """
        :return: Array of string, each of which is either "x" or is
                 composed of a hand ("<" or ">") identifief and a digit (1-5).

                 Returns None if any fingerings for the other hand
                 are detected.
        """
        if staff not in ("upper", "lower"):
            raise Exception("Invalid input: staff must be 'upper' or 'lower'.")

        ast = self.parse()
        if staff == "upper":
            lines = ast.upper
            hand = ">"
        else:
            lines = ast.lower
            hand = "<"

        handed_digits = []
        for line in lines:
            for score_fingering in line:
                strike = score_fingering.pf.fingering.strike
                if strike.hand and strike.hand != hand:
                    hand = strike.hand
                digit = strike.digit
                handed_digit = hand + str(digit)
                handed_digits.append(handed_digit)
        return handed_digits

    def __init__(self, abcdf=None, authority=None, authority_year=None, transcriber=None,
                 transcription_date=None, abcdf_id=None, comments=''):
        self._authority = authority
        self._authority_year = authority_year
        self._transcriber = transcriber
        self._transcription_date = transcription_date
        self._ast = None
        self._abcdf = None
        if abcdf:
            self.abcdf(abcdf)
        self._abcdf_id = abcdf_id
        self._comments = comments

    def authority(self, authority=None):
        if authority:
            self._authority = authority
        return self._authority

    def authority_year(self, authority_year=None):
        if authority_year:
            self._authority_year = authority_year
        return self._authority_year

    def transcriber(self, transcriber=None):
        if transcriber:
            self._transcriber = transcriber
        return self._transcriber

    def transcription_date(self, transcription_date=None):
        if transcription_date:
            self._transcription_date = transcription_date
        return self._transcription_date

    @staticmethod
    def _flatten(abcdf):
        flattened = abcdf.replace("&", "")
        return flattened

    def abcdf(self, abcdf=None, staff=None, flat=False):
        if abcdf:
            self._abcdf = abcdf
            self._ast = DAnnotation.ast_for_abcdf(abcdf)
        if staff == "upper":
            return self.upper_abcdf(flat=flat)
        elif staff == "lower":
            return self.lower_abcdf(flat=flat)
        if flat:
            return DAnnotation._flatten(self._abcdf)
        return self._abcdf

    def upper_abcdf(self, flat=False):
        (upper, lower) = self.abcdf().split('@')
        if flat:
            return DAnnotation._flatten(upper)
        return upper

    def lower_abcdf(self, flat=False):
        (upper, lower) = self.abcdf().split('@')
        if flat:
            return DAnnotation._flatten(lower)
        return lower

    def abcdf_id(self, abcdf_id=None):
        if abcdf_id:
            self._abcdf_id = abcdf_id
        return self._abcdf_id

    def comments(self, comments=None):
        if comments:
            self._comments = comments
        return self._comments.rstrip()

    def add_comment_line(self, comment):
        self._comments += comment + "\n"



#!/usr/bin/env python3
import re
import unittest
from pydactyl.dactyler.Hart import Hart
from pydactyl.dcorpus.DCorpus import DCorpus
import TestConstant


class HartTest(unittest.TestCase):
    @staticmethod
    def test_hart_edges():
        hart = Hart()
        d_corpus = DCorpus(corpus_str=TestConstant.ONE_NOTE)
        hart.load_corpus(d_corpus=d_corpus)
        upper_rh_advice = hart.advise(staff="upper")
        right_re = re.compile('^>\d$')
        assert right_re.match(upper_rh_advice), "Bad one-note, right-hand, upper-staff advice"
        # both_advice = hart.advise(staff="both")
        # both_re = re.compile('^>\d@$')
        # assert both_re.match(both_advice), "Bad one-note, segregated, both-staff advice"

        hart = Hart()
        d_corpus = DCorpus(corpus_str=TestConstant.ONE_BLACK_NOTE_PER_STAFF)
        hart.load_corpus(d_corpus=d_corpus)
        upper_advice = hart.advise(staff="upper")
        right_re = re.compile('^>2$')
        assert right_re.match(upper_advice), "Bad black-note, upper-staff advice"
        lower_advice = hart.advise(staff="lower")
        left_re = re.compile('^<2$')
        assert left_re.match(lower_advice), "Bad black-note, upper-staff advice"
        both_advice = hart.advise(staff="both")
        both_re = re.compile('^>2@<2$')
        assert both_re.match(both_advice), "Bad black-note, both-staff advice"
        lower_advice = hart.advise(staff="lower", first_digit=3)
        lower_re = re.compile('^<3$')
        assert lower_re.match(lower_advice), "Bad preset black-note, both-staff advice"

        hart = Hart()
        d_corpus = DCorpus(corpus_str=TestConstant.TWO_WHITE_NOTES_PER_STAFF)
        hart.load_corpus(d_corpus=d_corpus)
        upper_advice = hart.advise(staff="upper")
        right_re = re.compile('^>\d\d$')
        assert right_re.match(upper_advice), "Bad two white-note, upper-staff advice"
        upper_advice = hart.advise(staff="upper", first_digit=3, last_digit=2)
        right_re = re.compile('^>32$')
        assert right_re.match(upper_advice), "Bad preset two white-note, upper-staff advice"

    @staticmethod
    def test_pivot_alignment():
        hart = Hart()
        d_corpus = DCorpus(corpus_str=TestConstant.A_MAJ_SCALE_SHORT)
        hart.load_corpus(d_corpus=d_corpus)

        evaluations = hart.evaluate_pivot_alignment(staff="both")
        for ev in evaluations:
            print(ev)
        assert evaluations[0] > 0, "Undetected pivot alignment costs"
        assert evaluations[1] == 0, "Bad fish in pivot alignment barrel"

    @staticmethod
    def test_distance_metrics():
        hart = Hart()
        d_corpus = DCorpus(corpus_str=TestConstant.A_MAJ_SCALE)
        hart.load_corpus(d_corpus=d_corpus)
        complete_rh_advice = hart.advise(staff="upper")
        complete_rh_advice_len = len(complete_rh_advice)
        right_re = re.compile('^>\d+$')
        assert right_re.match(complete_rh_advice), "Bad right-hand, upper-staff advice"
        rh_advice = hart.advise(staff="upper", offset=3, first_digit=4)
        short_advice_len = len(rh_advice)
        assert complete_rh_advice_len - 3 == short_advice_len, "Bad offset for advise() call"
        ff_re = re.compile('^>4\d+$')
        assert ff_re.match(rh_advice), "Bad first finger constraint"

        rh_advice = hart.advise(staff="upper", offset=10, first_digit=5, last_digit=5)
        short_advice_len = len(rh_advice)
        assert complete_rh_advice_len - 10 == short_advice_len, "Bad offset for advise() call"
        ff_re = re.compile('^>5\d+5$')
        assert ff_re.match(rh_advice), "Bad first and last finger constraints"

        lh_advice = hart.advise(staff="lower")
        left_re = re.compile('^<\d+$')
        assert left_re.match(lh_advice), "Bad left-hand, lower-staff advice"
        combo_advice = hart.advise(staff="both")
        clean_combo_advice = re.sub('[><&]', '',  combo_advice)
        d_score = d_corpus.d_score_by_index(index=0)
        gold_fingering = d_score.abcdf(index=0)
        clean_gold_fingering = re.sub('[><&]', '',  gold_fingering)

        combo_re = re.compile('^>\d+@<\d+$')
        assert combo_re.match(combo_advice), "Bad combined advice"
        hamming_evaluations = hart.evaluate_strike_distance(method="hamming", staff="both")
        # for he in hamming_evaluations:
            # print(he)
        assert hamming_evaluations[0] > 0, "Undetected Hamming costs"
        assert hamming_evaluations[1] == 0, "Bad fish in Hamming barrel"

        natural_evaluations = hart.evaluate_strike_distance(method="natural", staff="both")
        # for he in natural_evaluations:
            # print(he)
        assert natural_evaluations[0] > 0, "Undetected natural costs"
        assert natural_evaluations[1] == 0, "Bad fish in natural barrel"

        pivot_evaluations = hart.evaluate_strike_distance(method="pivot", staff="both")
        # for he in pivot_evaluations:
            # print(he)
        assert pivot_evaluations[0] > 0, "Undetected pivot costs"
        assert pivot_evaluations[1] == 0, "Bad fish in pivot barrel"

    @staticmethod
    def test_reentry():
        hart = Hart()
        d_corpus = DCorpus(corpus_str=TestConstant.A_MAJ_SCALE_SHORT)
        hart.load_corpus(d_corpus=d_corpus)

        reentry_hamming_evals = hart.evaluate_strike_reentry(method="hamming", staff="upper", gold_indices=[0, 1])
        # for rhe in reentry_hamming_evals:
            # print("RHE:{0}".format(rhe))
        assert reentry_hamming_evals[0] > 0, "Undetected Hamming reentry costs"
        assert reentry_hamming_evals[1] == 0, "Bad fish in Hamming reentry barrel"

        reentry_hamming_evals = hart.evaluate_strike_reentry(method="hamming", staff="both")
        # for rhe in reentry_hamming_evals:
            # print("RHE:{0}".format(rhe))
        assert reentry_hamming_evals[0] > 0, "Undetected Hamming reentry costs"
        assert reentry_hamming_evals[1] == 0, "Bad fish in Hamming reentry barrel"
        hamming_score = reentry_hamming_evals[0]

        reentry_natural_evals = hart.evaluate_strike_reentry(method="natural", staff="both")
        # for rne in reentry_natural_evals:
            # print("RNE:{0}".format(rne))
        assert reentry_natural_evals[0] > 0, "Undetected natural reentry costs"
        assert reentry_natural_evals[1] == 0, "Bad fish in natural reentry barrel"
        natural_score = reentry_natural_evals[0]
        assert natural_score > hamming_score, "Reentry: Natural <= Hamming"

        reentry_pivot_evals = hart.evaluate_strike_reentry(method="pivot", staff="both")
        # for rpe in reentry_pivot_evals:
            # print("RPE:{0}".format(rpe))
        assert reentry_pivot_evals[0] > 0, "Undetected pivot reentry costs"
        assert reentry_pivot_evals[1] == 0, "Bad fish in pivot reentry barrel"
        pivot_score = reentry_pivot_evals[0]
        assert natural_score < pivot_score, "Reentry: Natural >= Pivot"

if __name__ == "__main__":
    unittest.main()  # run all tests



__author__ = 'David Randolph'
# Copyright (c) 2014-2018 David A. Randolph.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.


class DEvaluation:
    """
    Class to keep track of how well each of the top ranked outputs of a model
    performed against a given gold standard.
    """
    def __init__(self, gold_d_annotation, scores=[]):
        self._gold_annotation = gold_d_annotation
        self._scores = list()
        self.scores(scores)

    def gold_d_annotation(self, annotation=None):
        if annotation:
            self._gold_annotation = annotation
        return self._gold_annotation

    def scores(self, scores=[]):
        if scores:
            for score in scores:
                self._scores.append(score)
        return self._scores

    def append_score(self, score):
        if score:
            self._scores.append(score)
            return True
        return False

    def score(self, index=None):
        if index is None:
            index = 0
        if len(self._scores) > 0 and index < len(self._scores):
            return self._scores[index]
        return None


class DEvaluationSet:
    def __init__(self, d_score, gold_indices=[]):
        self._gold_by_index = dict()
        self._gold_by_id = dict()
        abcd_header = d_score.abcd_header()
        gold_index = 0
        for annot in abcd_header.annotations():
            if gold_indices and gold_index not in gold_indices:
                continue
            evaluation = DEvaluation(gold_d_annotation=annot)
            self._gold_by_index[gold_index] = evaluation
            gold_id = annot.abcdf_id()
            self._gold_by_id[gold_id] = evaluation

    def evaluation(self, gold_index=None, gold_id=None):
        if gold_index in self._gold_by_index:
            return self._gold_by_index[gold_index]
        elif gold_id in self._gold_by_id:
            return self._gold_by_id[gold_id]
        return None

    def scores(self, gold_index=None, gold_id=None, scores=[]):
        evaluation = self.evaluation(gold_index=gold_index, gold_id=gold_id)
        if evaluation is None:
            raise Exception("No DEvaluation found")
        if scores:
            evaluation.scores(scores=scores)
        return evaluation.scores()



__author__ = 'David Randolph'
# Copyright (c) 2018 David A. Randolph.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
""" Here we migrate a method originally presented for guitar to piano: 

        S. I. Sayegh, “Fingering for string instruments with the optimum
            path paradigm,” Comput. Music J., vol. 13, no. 3, pp. 76–84, 1989.
            
    A nugget cuz we dug it.
"""

import networkx as nx
import copy
from . import Dactyler as D
from pydactyl.dcorpus.DNote import DNote


class Sayegh(D.TrainedDactyler):

    def __init__(self, segmenter=None, segment_combiner="normal", staff_combiner="naive"):
        super().__init__(segmenter=segmenter, segment_combiner=segment_combiner, staff_combiner=staff_combiner)
        self._training = dict()  # W' in the Sayegh paper.

    @staticmethod
    def _learn_from_example(staff, transition_counts, fingered_counts, training):
        for (midi_1, midi_2, digit_1, digit_2) in fingered_counts:
            fingered_count = fingered_counts[(midi_1, midi_2, digit_1, digit_2)]
            transition_count = transition_counts[(midi_1, midi_2)]
            training[(midi_1, midi_2, digit_1, digit_2)] += fingered_count / transition_count;

    def _paths(self, staff):
        path_exists = dict()
        for (from_midi, to_midi, from_finger, to_finger) in self._training[staff]:
            weight = self._training[staff][(from_midi, to_midi, from_finger, to_finger)]
            if weight > 0:
                path_exists[(from_midi, to_midi)] = True
        return path_exists

    def train(self, d_corpus, staff="both", segregate=True, segmenter=None, annotation_indices=[]):
        if not segregate:
            raise Exception("Desegregated (integrated) fingering not supported.")

        if staff == "both":
            self.train(d_corpus, staff="upper", segmenter=segmenter, annotation_indices=annotation_indices)
            self.train(d_corpus, staff="lower", segmenter=segmenter, annotation_indices=annotation_indices)
            return

        training = dict()
        lo, hi = d_corpus.pitch_range(staff=staff)
        hand = ">"
        if staff == "lower":
            hand = "<"
        for from_midi in range(lo, hi + 1):
            for from_digit in range(1, 6):
                from_finger = hand + str(from_digit)
                for to_midi in range(lo, hi + 1):
                    for to_digit in range(1, 6):
                        to_finger = hand + str(to_digit)
                        training[(from_midi, to_midi, from_finger, to_finger)] = 0

        score_count = d_corpus.score_count()
        for i in range(score_count):
            d_score = d_corpus.d_score_by_index(i)
            self.squawk("Training from {0} score".format(d_score.title()))
            if d_score.is_fully_annotated(indices=annotation_indices):
                hdr = d_score.abcd_header()
                annot_index = 0
                for annot in hdr.annotations():
                    if len(annotation_indices) > 0 and annot_index not in annotation_indices:
                        continue
                    annot_index += 1

                    prior_midi = None
                    prior_digit = None
                    transition_counts = {}
                    fingered_counts = {}
                    handed_digits = annot.handed_strike_digits(staff=staff)
                    stave_notes = d_score.orderly_note_stream(staff=staff)
                    stave_segments = d_score.orderly_note_stream_segments(staff=staff)
                    if len(handed_digits) != len(stave_notes):
                        raise Exception("Strike digit mismatch in {0}: {1} notes and {2} digits".format(
                            d_score.title(), len(stave_notes), len(handed_digits)))

                    sf_count = annot.score_fingering_count(staff=staff)
                    segment_index = 0
                    note_index = 0
                    segment = stave_segments[segment_index]
                    for j in range(sf_count):
                        if note_index == len(segment):
                            Sayegh._learn_from_example(staff=staff,
                                                       transition_counts=transition_counts,
                                                       fingered_counts=fingered_counts,
                                                       training=training)
                            segment_index += 1
                            note_index = 0
                        current_note = stave_segments[segment_index][note_index]
                        current_midi = current_note.pitch.midi
                        current_digit = handed_digits[j]
                        if prior_midi is None:
                            prior_midi = current_midi
                            prior_digit = current_digit
                            note_index += 1
                            continue
                        self.squawk("{0} {1} --> {2} {3}".format(prior_midi, prior_digit, current_midi, current_digit))
                        if (prior_midi, current_midi) in transition_counts:
                            transition_counts[(prior_midi, current_midi)] += 1
                        else:
                            transition_counts[(prior_midi, current_midi)] = 1
                        if (prior_midi, prior_digit, current_midi, current_digit) in fingered_counts:
                            fingered_counts[(prior_midi, current_midi, prior_digit, current_digit)] += 1
                        else:
                            fingered_counts[(prior_midi, current_midi, prior_digit, current_digit)] = 1

                        prior_midi = current_midi
                        prior_digit = current_digit
                        note_index += 1

                    Sayegh._learn_from_example(staff=staff,
                                               transition_counts=transition_counts,
                                               fingered_counts=fingered_counts,
                                               training=training)
        self._training[staff] = training

    def generate_segment_advice(self, segment, staff, offset=0, cycle=False,
                                handed_first_digit=None, handed_last_digit=None, k=None):
        """
        Generate a set of k ranked fingering suggestions for the given segment.
        :param segment: The segment to work with, as a music21 score object.
        :param staff: The staff (one of "upper" or "lower") from which the segment was derived.
        :param offset: The zero-based index to begin the returned advice. Defaults to 0. FIXME: Not supported.
        :param handed_first_digit: Constrain the solution to begin with this finger.
        :param handed_last_digit: Constrain the solution to end with this finger.
        :param k: The number of advice segments to return. The actual number returned may be less,
        but will be no more, than this number.
        :return: suggestions, costs, details: Two lists are returned. The first contains suggested fingering
        solutions as abcDF strings. The second contains the respective costs of each suggestion. The third
        is a list of lists containing the specific cost of each fingering transition to the total suggestion cost.
        """
        if len(segment) == 1:
            note_list = DNote.note_list(segment)
            abcdf = D.Dactyler.one_note_advise(note_list[0], staff=staff,
                                               first_digit=handed_first_digit,
                                               last_digit=handed_last_digit)
            return [abcdf], [0], [0]

        hand = ">"
        if staff == "lower":
            hand = "<"

        g = nx.DiGraph()
        g.add_node(0, midi=None, handed_digit=None)
        prior_slice_node_ids = list()
        prior_slice_node_ids.append(0)
        last_note_in_segment_index = len(segment) - 1
        note_in_segment_index = 0
        node_id = 1
        on_last_prefingered_note = False
        for note in segment:
            on_first_prefingered_note = False
            slice_node_ids = list()

            if note_in_segment_index == 0 and handed_first_digit:
                on_first_prefingered_note = True

            if note_in_segment_index == last_note_in_segment_index and handed_last_digit:
                on_last_prefingered_note = True

            for digit in range(1, 6):
                handed_digit = hand + str(digit)
                if on_last_prefingered_note and handed_digit != handed_last_digit:
                    continue
                if on_first_prefingered_note and handed_digit != handed_first_digit:
                    continue
                g.add_node(node_id, midi=note.pitch.midi, handed_digit=handed_digit)
                slice_node_ids.append(node_id)
                if 0 in prior_slice_node_ids:
                    g.add_edge(0, node_id, weight=-1)
                else:
                    for prior_node_id in prior_slice_node_ids:
                        prior_node = g.nodes[prior_node_id]
                        prior_midi = prior_node["midi"]
                        prior_hd = prior_node["handed_digit"]
                        weight_key = (prior_midi, note.pitch.midi, prior_hd, handed_digit)
                        # We will be searching for the shortest path over DAG, so the
                        # weights of favorable steps must be made small (negative).
                        weight = -1 * self._training[staff][weight_key]
                        g.add_edge(prior_node_id, node_id, weight=weight)
                node_id += 1
            if len(slice_node_ids) > 0:
                prior_slice_node_ids = copy.copy(slice_node_ids)
            note_in_segment_index += 1

        g.add_node(node_id, midi=None, handed_digit=None)
        for prior_node_id in prior_slice_node_ids:
            g.add_edge(prior_node_id, node_id, weight=-1)

        return D.Dactyler.generate_standard_graph_advice(g=g, target_id=node_id, k=k)



__author__ = 'David Randolph'
# Copyright (c) 2014-2018 David A. Randolph.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
from abc import ABC, abstractmethod
import pickle
import re
import copy
import networkx as nx
import json
from itertools import islice
from datetime import datetime
from pydactyl.dactyler import Constant
from pydactyl.dcorpus.DCorpus import DCorpus
from pydactyl.dcorpus.DNote import AnnotatedDNote
from pydactyl.dcorpus.DAnnotation import DAnnotation
import os


class Dactyler(ABC):
    """Base class for all fingering algorithms."""

    # FIXME: The log should be timestamped and for the specific algorithm being used.
    SQUAWK_OUT_LOUD = False
    DELETE_LOG = True

    def __init__(self, segmenter=None, segment_combiner="normal", staff_combiner="naive"):
        self._d_corpus = None
        timestamp = datetime.now().isoformat()
        self._log_file_path = '/tmp/dactyler_' + self.__class__.__name__ + '_' + timestamp + '.log'
        self._log = open(self._log_file_path, 'a')
        self._segmenter = segmenter
        # One of "cost," "normal," or "rank"
        self._segment_combiner = segment_combiner

        self._staff_combiner = staff_combiner

    def __del__(self):
        self._log.close()
        if Dactyler.DELETE_LOG:
            os.remove(self._log_file_path)

    def segmenter(self, segmenter=None):
        if segmenter is None:
            return self._segmenter
        self._segmenter = segmenter

    def segment_combiner(self, method=None):
        if method is None:
            return self._segment_combiner
        self._segment_combiner = method

    def staff_combiner(self, method=None):
        if method is None:
            return self._staff_combiner
        self._staff_combiner = method


    @staticmethod
    def graphmlize(g):
        """
        Return a version of networkx graph g that can be written as a graphml file.
        :param g: A networkx graph.
        :return: h: A networkx graph with all scalar edge attributes.
        """
        h = g.copy()
        for u, v, d in h.edges(data=True):
            for key, val in d.items():
                if type(val) in (dict, tuple, list):
                    json_str = json.JSONEncoder().encode(val)
                    d[key] = json_str
        return h


    @staticmethod
    def combine_abcdf_segments(segments):
        """
        Combine the array of abcDF strings into a single, simplified abcDF string.
        """
        abcdf = ''
        current_hand = None
        for seg in segments:
            for ch in seg:
                if ch == "<" or ch == ">":
                    if ch != current_hand:
                        abcdf += ch
                        current_hand = ch
                else:
                    abcdf += ch
        return abcdf

    def combine_staves(self, upper_suggestions, upper_costs, upper_details, upper_length,
                       lower_suggestions, lower_costs, lower_details, lower_length, k=1):
        """
        Apply the staff_combiner to combine solutions for upper and lower staves.
        :param upper_suggestions:
        :param upper_costs:
        :param lower_suggestions:
        :param lower_costs:
        :param upper_length:
        :param lower_length:
        :param k:
        :return:
        """
        # FIXME
        suggestions = list()
        costs = list()
        details = list()
        if self.staff_combiner() == "naive":
            if len(upper_suggestions) == len(lower_suggestions):
                for i in range(len(upper_suggestions)):
                    suggestions.append(upper_suggestions[i] + "@" + lower_suggestions[i])
                    costs.append(upper_costs[i] + lower_costs[i])
            elif len(upper_suggestions) < len(lower_suggestions):
                for lower_index in range(len(lower_suggestions)):
                    upper_index = lower_index // len(lower_suggestions)
                    suggestions.append(upper_suggestions[upper_index] + "@" + lower_suggestions[lower_index])
                    costs.append(upper_costs[upper_index] + lower_costs[lower_index])
                    details.extend(upper_details[upper_index])
                    details.extend(lower_details[lower_index])
            else:
                for upper_index in range(len(upper_suggestions)):
                    lower_index = upper_index // len(upper_suggestions)
                    suggestions.append(upper_suggestions[upper_index] + "@" + lower_suggestions[lower_index])
                    costs.append(upper_costs[upper_index] + lower_costs[lower_index])
                    details.extend(upper_details[upper_index])
                    details.extend(lower_details[lower_index])
        return suggestions, costs, details

    def combine_segments(self, suggestions_for_segment, costs_for_segment, details_for_segment, segment_lengths, k=1):
        """
        Given a list of fingering suggestions for an ordered list of fragments in an entire score,
        combine them in the k-best ways. This reduces to a k-shortest path search problem, where the arc costs
        are simply the costs associated with the suggested segment fingering. How these segment fingering costs
        are calculated affects the rankings of the complete end-to-end suggestions returned. The simplest method is
        perhaps the truest--we just use the cost reported for the segment in question. The cost of the entire
        sequence should be the sum of the individual segment costs. However, this will make the most difficult and
        longest segment fingerings the most consistent in the returned solutions. These may be the very segments
        for which a user would appreciate more suggestions. These would also be the segments that would have less
        agreement among pianists. It might be better to normalize the cost by the number of notes in the segment
        or even to only look at the ordinal rank of segment suggestions, so we see a similar amount of variability
        for each segment. These three costing approaches are supported here, and must be specified in via the
        self.segment_combiner() as one of "cost," "normal," or "rank."
        :param suggestions_for_segment: A list of lists of suggested fingerings for each segment.
        :param costs_for_segment: A corresponding list of lists of costs for each suggestion.
        :param details_for_segment: A corresponding list of data structures detailing how the cost
        for each suggestion was determined.
        :param segment_lengths: A corresponding list of lists of the lengths of each segment. Used to
        normalize the contribution of a segment's contribution to the total cost.
        :param k: The number of advice segments to return. The actual number returned may be less,
        but will be no more, than this number.
        :return: suggestions, costs, details: Three lists are returned. The first contains suggested fingering
        solutions as abcDF strings. The second contains the respective costs of each complete suggestion, adjusted
        per self.segment_combiner(). The third contains the pre-adjusted cost details for each suggestion.
        (The format of these details varies by algorithm used.)
        """
        g = nx.DiGraph()
        g.add_node(0, suggestion=None)
        prior_slice_node_ids = list()
        prior_slice_node_ids.append(0)
        node_id = 1
        for segment_index in range(len(suggestions_for_segment)):
            slice_node_ids = list()
            suggestions = suggestions_for_segment[segment_index]
            costs = costs_for_segment[segment_index]
            details = details_for_segment[segment_index]
            for suggestion_id in range(len(suggestions)):
                g.add_node(node_id, suggestion=suggestions[suggestion_id])
                for prior_node_id in prior_slice_node_ids:
                    if self.segment_combiner() == 'cost':
                        cost = costs[suggestion_id]
                    elif self.segment_combiner() == 'normal':
                        cost = costs[suggestion_id] / segment_lengths[segment_index]
                    elif self.segment_combiner() == 'rank':
                        cost = suggestion_id
                    else:
                        raise Exception("Unsupported method: {0}".format(self.segment_combiner()))
                    g.add_edge(prior_node_id, node_id, weight=cost, details=details[suggestion_id])
                slice_node_ids.append(node_id)
                node_id += 1
            if len(slice_node_ids) > 0:
                prior_slice_node_ids = copy.copy(slice_node_ids)

        g.add_node(node_id, suggestion=None)
        for prior_node_id in prior_slice_node_ids:
            g.add_edge(prior_node_id, node_id, weight=0)

        if k is None or k == 1:
            path = nx.shortest_path(g, source=0, target=node_id, weight="weight")
            abcdf_segments = list()
            for node_id in path:
                node = g.nodes[node_id]
                if node["suggestion"]:
                    abcdf_segments.append(node["suggestion"])
            suggestion = Dactyler.combine_abcdf_segments(abcdf_segments)
            cost = nx.shortest_path_length(g, source=0, target=node_id, weight="weight")
            sub_g = g.subgraph(path)
            suggestion_details = list()
            for (u, v, segment_details) in sub_g.edges.data('details'):
                if not segment_details:
                    continue
                suggestion_details.append(segment_details)
            return [suggestion], [cost], [suggestion_details]
        else:
            suggestions = list()
            costs = list()
            details = list()
            k_best_paths = list(islice(nx.shortest_simple_paths(g, source=0, target=node_id, weight="weight"), k))
            for path in k_best_paths:
                sub_g = g.subgraph(path)
                cost = sub_g.size(weight="weight")
                abcdf_segments = list()
                for node_id in path:
                    node = g.nodes[node_id]
                    if node["suggestion"]:
                        abcdf_segments.append(node["suggestion"])
                suggestion = Dactyler.combine_abcdf_segments(abcdf_segments)
                suggestions.append(suggestion)
                costs.append(cost)
                suggestion_details = list()
                for (u, v, segment_details) in sub_g.edges.data('details'):
                    if not segment_details:
                        continue
                    suggestion_details.append(segment_details)
                details.append(suggestion_details)
            return suggestions, costs, details

    @staticmethod
    def hand_digit(digit, staff):
        """
        Determine the handed digit for the input digit string.
        If the string already has an assigned digit, just return it.
        Otherwise assign the default hand for the specified staff.
        :return: The abcDF hand digit (">2" or "<3").
        """
        if digit is None:
            return None

        handed_re = re.compile('^[<>]\d$')
        if handed_re.match(str(digit)):
            return digit

        staff_prefix = ">"
        if staff == "lower":
            staff_prefix = "<"
        handed_digit = staff_prefix + str(digit)
        return handed_digit

    @staticmethod
    def digit_hand(handed_digit):
        """
        Determine the hand specified in the handed digit.
        :return: The abcDF hand specifier (">" or "<").
        """
        handed_re = re.compile('^([<>]{1})\d$')
        mat = handed_re.match(str(handed_digit))
        hand = mat.group(1)
        if hand != "<" and hand != ">":
            raise Exception("Ill-formed handed digit: {0}".format(handed_digit))
        return hand

    def squawk(self, msg):
        self._log.write(str(msg) + "\n")
        if Dactyler.SQUAWK_OUT_LOUD:
            print(str(msg))

    def squeak(self, msg):
        self._log.write(str(msg))
        if Dactyler.SQUAWK_OUT_LOUD:
            print(str(msg), end="")

    @staticmethod
    def one_note_advise(d_note, staff="upper", first_digit=None, last_digit=None):
        """
        Dispense advice for the degenerate case where we have only one note to consider in isolation.
        Unless otherwise constrained, we prefer the thumb for a white key and an index finger
        for a black key.
        :param d_note: The note to consider.
        :param staff: The staff containing the note.
        :param first_digit: The digit to use is constrained to be this.
        :param last_digit: The digit is constrained to be this.
        :return: Advice in the form of a handed digit (e.g., ">2").
        """
        if staff != "upper" and staff != "lower":
            raise Exception("One note advice not available for {0} staff.".format(staff))
        if first_digit and last_digit and first_digit != last_digit:
            raise Exception("Ambiguous digit constraint: {0} and {1}".format(first_digit, last_digit))

        if staff == "upper":
            advice = ">"
        else:
            advice = "<"

        digit = "1"
        if first_digit:
            digit = str(first_digit)
        elif last_digit:
            digit = last_digit
        elif d_note.is_black():
            digit = "2"
        advice += digit

        return advice

    @abstractmethod
    def generate_segment_advice(self, segment, staff, offset=0, cycle=None,
                                handed_first_digit=None, handed_last_digit=None, k=1):
        """
        Abstract method must be defined by derived class.
        Generate a set of k ranked fingering suggestions for the given segment.
        :param segment: The segment to work with, as a music21 score object.
        :param staff: The staff (one of "upper" or "lower") from which the segment was derived.
        :param offset: The zero-based index to begin the returned advice.
        :param cycle: Detect repeating note patterns of at least this length within each segment and generate
        advice best suited for uniform fingerings of the repeated patterns. Defaults to None (ignore cycles).
        :param handed_first_digit: Constrain the solution to begin with this finger.
        :param handed_last_digit: Constrain the solution to end with this finger.
        :param k: The number of advice segments to return. The actual number returned may be less,
        but will be no more, than this number.
        :return: suggestions, costs, details: Three lists are returned. The first contains suggested fingering
        solutions as abcDF strings. The second contains the respective costs of each suggestion. The third
        contains details about how each cost was determined.
        """
        return list(), list(), list()

    @staticmethod
    def generate_standard_graph_advice(g, target_id, k):
        """
        Apply standard shortest path algorithms to determine set of optimal fingerings based on
        a standardized networkx graph.
        :param g: The weighted graph. Weights must be specified via a "weight" edge parameter. Fingerings
        must be set on each "handed_digit" node parameter.
        :param target_id: The node id (key) for the last node or end point in the graph.
        :param k: The number of
        :return: suggestions, costs: Two lists are returned. The first contains suggested fingering
        solutions as abcDF strings. The second contains the respective costs of each suggestion.  The third
        is a list of lists containing the specific cost of each fingering transition to the total suggestion cost.
        """
        if k is None or k == 1:
            path = nx.shortest_path(g, source=0, target=target_id, weight="weight")
            segment_abcdf = ''
            for node_id in path:
                node = g.nodes[node_id]
                if node["handed_digit"]:
                    segment_abcdf += node["handed_digit"]
            cost = nx.shortest_path_length(g, source=0, target=target_id, weight="weight")
            sub_g = g.subgraph(path)
            node_costs = list()
            for (u, v, weight) in sub_g.edges.data('weight'):
                if not weight:
                    continue
                node_costs.append(weight)
            return [segment_abcdf], [cost], [node_costs]
        else:
            sugg_map = dict()
            suggestions = list()
            costs = list()
            details = list()
            k_best_paths = list(islice(nx.shortest_simple_paths(g, source=0, target=target_id, weight="weight"), k))
            for path in k_best_paths:
                sub_g = g.subgraph(path)
                suggestion_cost = sub_g.size(weight="weight")
                segment_abcdf = ''
                for node_id in path:
                    node = g.nodes[node_id]
                    if node["handed_digit"]:
                        segment_abcdf += node["handed_digit"]
                suggestions.append(segment_abcdf)
                if segment_abcdf in sugg_map:
                    sugg_map[segment_abcdf] += 1
                else:
                    sugg_map[segment_abcdf] = 1
                costs.append(suggestion_cost)

                node_costs = list()
                for (u, v, weight) in sub_g.edges.data('weight'):
                    if not weight:
                        continue
                    node_costs.append(weight)
                details.append(node_costs)
            print("TOTAL: {0} DISTINCT: {1} COSTS: {2}".format(len(suggestions), len(sugg_map), costs))
            return suggestions, costs, details

    def generate_advice(self, score_index=0, staff="upper", cycle=None, offset=0, first_digit=None, last_digit=None, k=1):
        """
        Generate advice for the specified score. This method only supports segregated advice.
        :param score_index:
        :param staff: One of "upper," "lower," or "both." Note that requesting "both" staves will result in
        the pairwise combination of k suggestions from upper and lower staves and the conflation of the staff
        costs if the default "naive" staff_combiner() is used. We only guarantees that the first suggestion
        returned is globally optimal. Additional combination methods may be provided, but the caller should
        generate advice for upper and lower staves separately to be able to combine advice more intelligently.
        :param cycle: Detect repeating note patterns of at least this length within each segment and generate
        advice best suited for uniform fingerings of the repeated patterns. Defaults to None (ignore cycles).
        :param offset:
        :param first_digit: Constrain the advice to begin with this digit (1-5). Inconsistent with "both"
        staff parameter.
        :param last_digit: Constrain the advice to end with this digit (1-5). Inconsistent with "both"
        staff parameter.
        :param k: The number of suggestions and corresponding costs to return. (Up to this number may be
        returned.)
        :return: suggestions, costs: Two lists are returned. The first contains suggested fingering
        solutions as abcDF strings. The second contains the respective costs of each suggestion.
        """
        d_scores = self._d_corpus.d_score_list()
        if score_index >= len(d_scores):
            raise Exception("Score index out of range")

        d_score = d_scores[score_index]
        if staff == "both":
            if d_score.part_count() < 2:
                raise Exception("Only one staff present.")

            if offset or first_digit or last_digit:
                raise Exception("Ambiguous use to offset and/or first/last digit for both staves.")

            upper_suggestions, upper_costs, upper_details = self.generate_advice(
                score_index=score_index, staff="upper", cycle=cycle, k=k)
            lower_suggestions, lower_costs, lower_details = self.generate_advice(
                score_index=score_index, staff="lower", cycle=cycle, k=k)
            upper_length = d_score.upper_d_part().length()
            lower_length = d_score.lower_d_part().length()
            return self.combine_staves(
                upper_suggestions=upper_suggestions, upper_costs=upper_costs, upper_details=upper_details,
                lower_suggestions=lower_suggestions, lower_costs=lower_costs, lower_details=lower_details,
                upper_length=upper_length, lower_length=lower_length, k=k)

        if staff != "upper" and staff != "lower":
            raise Exception("Segregated advice is only dispensed one staff at a time.")

        handed_first_digit = Dactyler.hand_digit(digit=first_digit, staff=staff)
        handed_last_digit = Dactyler.hand_digit(digit=last_digit, staff=staff)

        if d_score.part_count() == 1:
            d_part = d_score.combined_d_part()
        else:
            # We support (segregated) left hand fingerings. By segregated, we
            # mean the right hand is dedicated to the upper staff, and the
            # left hand is dedicated to the lower staff.
            d_part = d_score.d_part(staff=staff)

        segments = d_part.orderly_note_stream_segments(offset=offset)
        segment_index = 0
        last_segment_index = len(segments) - 1
        segment_lengths = list()
        suggestions_for_segment = list()
        costs_for_segment = list()
        details_for_segment = list()
        for segment in segments:
            segment_offset = 0
            segment_handed_first = None
            segment_handed_last = None
            if segment_index == 0:
                segment_offset = offset
                segment_handed_first = handed_first_digit
            segment_length = len(segment) - segment_offset
            if segment_index == last_segment_index:
                segment_handed_last = handed_last_digit

            suggestions, costs, details = self.generate_segment_advice(segment=segment, staff=staff,
                                                                       offset=segment_offset,
                                                                       cycle=cycle,
                                                                       handed_first_digit=segment_handed_first,
                                                                       handed_last_digit=segment_handed_last, k=k)
            suggestions_for_segment.append(suggestions)
            costs_for_segment.append(costs)
            details_for_segment.append(details)
            segment_lengths.append(segment_length)
            segment_index += 1

        suggestions, costs, details = self.combine_segments(suggestions_for_segment=suggestions_for_segment,
                                                            costs_for_segment=costs_for_segment,
                                                            details_for_segment=details_for_segment,
                                                            segment_lengths=segment_lengths, k=k)
        return suggestions, costs, details

    @staticmethod
    def report_on_advice(suggestions, costs, details):
        print("")
        for i in range(len(suggestions)):
            print("{0}) {1}\ncost: {2} segments: {3}".format(i+1, suggestions[i], costs[i], len(details[i])))
            for j in range(len(details[i])):
                print(details[i][j])
            print("")
        print("")

    def advise(self, score_index=0, staff="upper", offset=0, first_digit=None, last_digit=None):
        """
        Generate advice for the specified score. This method only supports segregated advice.
        :param score_index:
        :param staff: One of "upper," "lower," or "both."
        :param offset: The zero-based index from which to start the advice.
        :param first_digit: Constrain the advice to begin with this digit (1-5). Inconsistent with "both"
        staff parameter.
        :param last_digit: Constrain the advice to end with this digit (1-5). Inconsistent with "both"
        staff parameter.
        :param k: The number of suggestions and corresponding costs to return. (Up to this number may be
        returned.)
        :return: abcDF advice string.
        """
        suggestions, costs, details = self.generate_advice(score_index=score_index, staff=staff, offset=offset,
                                                           first_digit=first_digit, last_digit=last_digit)
        return suggestions[0]

    def load_corpus(self, d_corpus=None, path=None):
        """
        Load corpus to be processed by the Dactyler model.
        :param d_corpus: A DCorpus object.
        :param path: The path to a file that can be opened as a DCorpus object.
        :return:
        """
        if d_corpus:
            self._d_corpus = d_corpus
        elif path:
            self._d_corpus = DCorpus(path)
        else:
            raise Exception("No corpus specified for Dactyler.")

    @staticmethod
    def strike_distance_cost(gold_hand, gold_digit, test_hand, test_digit, method="hamming"):
        if method == "hamming":
            if test_digit != gold_digit or test_hand != gold_hand:
                return 1
            else:
                return 0

        one = str(gold_hand) + str(gold_digit)
        other = str(test_hand) + str(test_digit)
        if method == "natural":
            cost = Constant.NATURAL_EDIT_DISTANCES[(one, other)]
            return cost
        elif method == "pivot":
            cost = Constant.PIVOT_EDIT_DISTANCES[(one, other)]
            return cost
        else:
            raise Exception("Unsupported method: {0}".format(method))

    def score_note_count(self, score_index=0, staff="both"):
        d_score = self._d_corpus.d_score_by_index(score_index)
        note_count = d_score.note_count(staff=staff)
        return note_count

    @staticmethod
    def _distance_and_loc(method, staff, test_annot, gold_annot, gold_offset=0, zero_cost=False):
        current_gold_hand = ">" if staff == "upper" else "<"
        current_test_hand = ">" if staff == "upper" else "<"

        test_sf_count = test_annot.score_fingering_count(staff=staff)
        gold_sf_count = gold_annot.score_fingering_count(staff=staff)

        adjusted_gold_sf_count = gold_sf_count - gold_offset
        if test_sf_count != adjusted_gold_sf_count:
            raise Exception("Length mismatch: test: {0} gold: {1}".format(test_sf_count, adjusted_gold_sf_count))

        score = 0
        i = None
        gold_digit = None
        for i in range(test_sf_count):
            gold_i = i + gold_offset
            gold_sf = gold_annot.score_fingering_at_index(index=gold_i, staff=staff)
            gold_strike = gold_sf.pf.fingering.strike
            gold_hand = gold_strike.hand if gold_strike.hand else current_gold_hand
            gold_digit = int(gold_strike.digit)

            test_sf = test_annot.score_fingering_at_index(index=i, staff=staff)
            test_strike = test_sf.pf.fingering.strike
            test_hand = test_strike.hand if test_strike.hand else current_test_hand
            test_digit = int(test_strike.digit)

            current_gold_hand = gold_hand
            current_test_hand = test_hand

            cost = Dactyler.strike_distance_cost(method=method,
                                                 gold_hand=gold_hand,
                                                 gold_digit=gold_digit,
                                                 test_hand=test_hand,
                                                 test_digit=test_digit)
            if zero_cost and cost:
                return cost, gold_i, gold_digit
            score += cost

        return score, i, gold_digit

    @staticmethod
    def _eval_strike_distance(method, staff, test_annot, gold_annot):
        (cost, location, gold_digit) = Dactyler._distance_and_loc(method=method, staff=staff,
                                                                  test_annot=test_annot, gold_annot=gold_annot)
        return cost

    def evaluate_strike_distance(self, method="hamming", score_index=0, staff="upper", gold_indices=[]):
        """
        Evaluate the best solution for a given score reported by the model against each of the specified
        gold-standard annotations embedded in the DScore object using the specified method.
        :param method: The edit distance metric to apply in the evaluation. One of "hamming," "natural," or "pivot."
        :param score_index: The zero-based index of the DScore within the DCorpus currently loaded in this Dactyler.
        :param staff: The staff to include in the evaluation (one of "upper," "lower," or "both."
        :param gold_indices: The zero-based indices of the DAnnotation objects from the DScore's ABCDHeader to use
        as the gold standard advice. If not specified, distances against all DAnnotations is calculated.
        :return: An array of distance measures against each specified gold index, in sorted order.
        """
        d_score = self._d_corpus.d_score_by_index(score_index)
        if not d_score.is_fully_annotated():
            raise Exception("Only fully annotated scores can be evaluated.")

        if staff == "both":
            staves = ['upper', 'lower']
        else:
            staves = [staff]

        test_abcdf = self.advise(score_index=score_index, staff=staff)
        test_annot = DAnnotation(abcdf=test_abcdf)
        hdr = d_score.abcd_header()
        scores = []
        gold_index = 0
        for gold_annot in hdr.annotations():
            if gold_indices and gold_index not in gold_indices:
                gold_index += 1
                continue
            score = 0
            for staff in staves:
                score += Dactyler._eval_strike_distance(method=method, staff=staff,
                                                        test_annot=test_annot, gold_annot=gold_annot)
            scores.append(score)
            gold_index += 1

        return scores

    def evaluate_strike_distances(self, method="hamming", score_index=0, staff="upper", gold_indices=[], k=1):
        """
        Evaluate the k best solutions for a given score reported by the model against each of the specified
        gold-standard annotations embedded in the DScore object.
        :param method: The edit distance metric to apply in the evaluation. One of "hamming," "natural," or "pivot."
        :param score_index: The zero-based index of the DScore within the DCorpus currently loaded in this Dactyler.
        :param staff: The staff to include in the evaluation (one of "upper" or "lower.")
        :param gold_indices: The zero-based indices of the DAnnotation objects from the DScore's ABCDHeader to use
        as the gold standard advice.
        :param k: The number of "k-best" abcDF advice strings produced by the Dactyler model to evaluate.
        :return suggestions, costs, scores_for_gold_index: An array of the suggested abcDF strings. An array of
        the costs determined by the model, and a dictionary mapping gold-standard indices to an array of corresponding
        distance metric scores.
        """
        if staff != "upper" and staff != "lower":
            raise Exception("Strike distances must be evaluated one staff at a time.")

        d_score = self._d_corpus.d_score_by_index(score_index)
        if not d_score.is_fully_annotated():
            raise Exception("Only fully annotated scores can be evaluated.")

        suggestions, costs = self.generate_advice(score_index=score_index, staff=staff, k=k)
        hdr = d_score.abcd_header()
        scores_for_gold_index = dict()
        gold_index = 0

        for gold_annot in hdr.annotations():
            if gold_indices and gold_index not in gold_indices:
                gold_index += 1
                continue
            scores = []
            for i in range(len(suggestions)):
                test_annot = DAnnotation(abcdf=suggestions[i])
                score = Dactyler._eval_strike_distance(method=method, staff=staff,
                                                       test_annot=test_annot, gold_annot=gold_annot)
                scores.append(score)

            scores_for_gold_index[gold_index] = scores
            gold_index += 1

        return suggestions, costs, scores_for_gold_index

    def evaluate_strike_reentry(self, method="hamming", score_index=0, staff="upper", gold_indices=[]):
        d_score = self._d_corpus.d_score_by_index(score_index)
        if not d_score.is_fully_annotated():
            raise Exception("Only fully annotated scores can be evaluated.")

        if staff == "both":
            staves = ['upper', 'lower']
        else:
            staves = [staff]

        hdr = d_score.abcd_header()
        scores = {"upper": [], "lower": []}
        for staff in staves:
            current_gold_index = 0
            for gold_annot in hdr.annotations():
                score = 0
                if len(gold_indices) > 0 and current_gold_index not in gold_indices:
                    current_gold_index += 1
                    continue
                current_gold_index += 1

                test_abcdf = self.advise(score_index=score_index, staff=staff)
                self.squawk("COMPLETE ADVICE: {0}".format(test_abcdf))
                if staff == 'upper':
                    test_abcdf += '@'
                else:
                    test_abcdf = '@' + str(test_abcdf)
                test_annot = DAnnotation(abcdf=test_abcdf)
                (cost, loc, gold_digit) = self._distance_and_loc(zero_cost=True, method=method, staff=staff,
                                                                 test_annot=test_annot, gold_annot=gold_annot)
                score += cost
                self.squawk("GOLD: {0} staff for {1}".format(staff, gold_annot.abcdf(staff=staff, flat=True)))
                self.squawk("TEST: {0} staff for {1}".format(staff, test_abcdf))
                self.squawk("SCORE: {0} COST: {1} LOC: {2}".format(score, cost, loc))

                while cost > 0:
                    test_abcdf = self.advise(score_index=score_index, staff=staff,
                                             offset=loc, first_digit=gold_digit)
                    self.squawk("GOLD: {0} staff for {1}".format(staff, gold_annot.abcdf(staff=staff, flat=True)))
                    self.squawk("TRUNCATED ADVICE: {0}".format(test_abcdf))
                    if staff == 'upper':
                        test_abcdf += '@'
                    else:
                        test_abcdf = '@' + str(test_abcdf)
                    test_annot = DAnnotation(abcdf=test_abcdf)
                    (cost, loc, gold_digit) = self._distance_and_loc(zero_cost=True, gold_offset=loc,
                                                                     method=method, staff=staff,
                                                                     test_annot=test_annot, gold_annot=gold_annot)
                    score += cost
                    self.squawk("     score: {0} cost: {1} loc: {2}".format(score, cost, loc))

                scores[staff].append(score)

        total_scores = []
        if len(scores['upper']) > 0:
            for i in range(len(scores['upper'])):
                total_scores.append(scores['upper'][i])
                if len(scores['lower']) > 0:
                    total_scores[i] += scores['lower'][i]
        elif len(scores['lower']) > 0:
            for i in range(len(scores['lower'])):
                total_scores.append(scores['lower'][i])
        else:
            raise Exception("No scores found.")

        return total_scores

    @staticmethod
    def _pivot_flags(staff, d_notes, annot):
        current_hand = ">" if staff == "upper" else "<"

        sf_count = annot.score_fingering_count(staff=staff)

        pivot_flags = list()
        prior_ad_note = None
        for i in range(sf_count):
            d_note = d_notes[i]
            sf = annot.score_fingering_at_index(index=i, staff=staff)
            strike = sf.pf.fingering.strike
            hand = strike.hand if strike.hand else current_hand
            digit = int(strike.digit)
            ad_note = AnnotatedDNote(m21_note=d_note.m21_note(),
                                     prior_note=prior_ad_note,
                                     strike_hand=hand,
                                     strike_digit=digit)

            if ad_note.is_pivot():
                pivot_flags.append(True)
            else:
                pivot_flags.append(False)
            prior_ad_note = ad_note
            current_hand = hand

        return pivot_flags

    @staticmethod
    def _pivot_alignment_score(gold_pivot_flags, test_pivot_flags, cost=1):
        if len(gold_pivot_flags) != len(test_pivot_flags):
            raise Exception("Mismatched pivot flag lists.")

        score = 0
        for i in range(len(gold_pivot_flags)):
            if gold_pivot_flags[i] != test_pivot_flags[i]:
                score += cost
        return score

    def evaluate_pivot_alignment(self, score_index=0, staff="upper", gold_indices=[]):
        d_score = self._d_corpus.d_score_by_index(score_index)
        if not d_score.is_fully_annotated():
            raise Exception("Only fully annotated scores can be evaluated.")

        if staff == "both":
            staves = ['upper', 'lower']
        else:
            staves = [staff]

        test_abcdf = self.advise(score_index=score_index, staff=staff)
        test_annot = DAnnotation(abcdf=test_abcdf)
        hdr = d_score.abcd_header()
        scores = []
        gold_index = 0
        for gold_annot in hdr.annotations():
            if gold_indices and gold_index not in gold_indices:
                gold_index += 1
                continue
            score = 0
            for staff in staves:
                d_part = d_score.d_part(staff=staff)
                d_notes = d_part.orderly_d_notes()
                test_pivot_flags = Dactyler._pivot_flags(staff=staff, d_notes=d_notes, annot=test_annot)
                gold_pivot_flags = Dactyler._pivot_flags(staff=staff, d_notes=d_notes, annot=gold_annot)
                staff_score = self._pivot_alignment_score(gold_pivot_flags=gold_pivot_flags, test_pivot_flags=test_pivot_flags)
                score += staff_score
            scores.append(score)
            gold_index += 1

        return scores


class TrainedDactyler(Dactyler):
    def __init__(self):
        super().__init__()
        self._training = {}

    @abstractmethod
    def generate_segment_advice(self, segment, staff, offset=0, cycle=None,
                                handed_first_digit=None, handed_last_digit=None, k=None):
        """
        Abstract method must be defined by derived class.
        Generate a set of k ranked fingering suggestions for the given segment.
        :param segment: The segment to work with, as a music21 score object.
        :param staff: The staff (one of "upper" or "lower") from which the segment was derived.
        :param offset: The zero-based index to begin the returned advice.
        :param cycle: Detect repeating note patterns of at least this length within each segment and generate
        advice best suited for uniform fingerings of the repeated patterns. Defaults to None (ignore cycles).
        :param handed_first_digit: Constrain the solution to begin with this finger.
        :param handed_last_digit: Constrain the solution to end with this finger.
        :param k: The number of advice segments to return. The actual number returned may be less,
        but will be no more, than this number.
        :return: suggestions, costs, details: Three lists are returned. The first contains suggested fingering
        solutions as abcDF strings. The second contains the respective costs of each suggestion. The third
        contains details about how each cost was determined.
        """
        pass

    @abstractmethod
    def train(self, d_corpus, staff="both", segregate=True, segmenter=None, annotation_indices=[]):
        return

    def retain(self, pickle_path=None, to_db=False):
        if pickle_path:
            with open(pickle_path, 'wb') as pickle_file:
                pickle.dump(self._training, pickle_file, pickle.HIGHEST_PROTOCOL)
        elif to_db:
            raise Exception("Retaining pickled file to database not yet supported.")
        else:
            raise Exception("No retention destination specified.")

    def recall(self, pickle_path=None, pickle_db_id=None):
        if pickle_path:
            with open(pickle_path, 'rb') as pickle_file:
                self._training = pickle.load(pickle_file)
        elif pickle_db_id:
            raise Exception("Recalling pickled training from database not yet supported.")
        else:
            raise Exception("No source specified from which to recall training.")

    def training(self):
        return self._training

    def demonstrate(self):
        for k in self._training:
            print(k, ': ', self._training[k])




#!/usr/bin/env python3
import re
import unittest
from pydactyl.dactyler.Hart import Hart
from pydactyl.dactyler.Hart import HartK
from pydactyl.dcorpus.DCorpus import DCorpus
import TestConstant


class RankTest(unittest.TestCase):
    @staticmethod
    def test_hartk():
        hart = Hart()
        d_corpus = DCorpus(corpus_str=TestConstant.A_MAJ_SCALE_SHORT)
        hart.load_corpus(d_corpus=d_corpus)
        hart_rh_advice = hart.advise(staff="upper")
        print(hart_rh_advice)

        hart_lh_advice = hart.advise(staff="lower")
        print(hart_lh_advice)

        hart_k = HartK()
        hart_k.load_corpus(d_corpus=d_corpus)
        hart_k.segment_combination_method("cost")

        suggestions, costs, hds_for_gold_index = hart_k.evaluate_strike_distances(method="hamming", staff="upper", k=20)
        min_cost = None
        for i in range(len(suggestions)):
            print("{0}:::{1}".format(costs[i], suggestions[i]))
            if suggestions[i] == hart_rh_advice:
                print("GOT IT!")
                break

        for gi in hds_for_gold_index:
            print("***" + str(gi) + "***")
            hd_count = len(hds_for_gold_index[gi])
            print("HD count for GI {0} = {1}".format(gi, hd_count))
            for hdi in range(hd_count):
                if hds_for_gold_index[gi][hdi] == 0:
                    print("GOT A ZERO HAMMING DISTANCE")

        # assert hamming_evaluations[0] > 0, "Undetected Hamming costs"
        # assert hamming_evaluations[1] == 0, "Bad fish in Hamming barrel"


if __name__ == "__main__":
    unittest.main()  # run all tests



#!/usr/bin/env python3
__author__ = 'David Randolph'
# Copyright (c) 2018 David A. Randolph.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
import re
import unittest
import TestConstant
from pydactyl.dcorpus.DCorpus import DCorpus
from pydactyl.dcorpus.DPart import DPart

BERINGER2_SCALE_CORPUS = TestConstant.BERINGER2_DIR + "/scales.abc"


class DCorpusTest(unittest.TestCase):
    @staticmethod
    def test_abc_2part_mono():
        d_corpus = DCorpus(corpus_path=BERINGER2_SCALE_CORPUS)
        lo, hi = d_corpus.pitch_range()
        assert lo < hi, "Bad dcorpus pitch range"
        score_count = d_corpus.score_count()
        assert score_count == 38, "Bad score count"
        d_score = d_corpus.d_score_by_title("scales_g_minor_melodic")
        assert d_score is not None, "Bad score_by_title retrieval"
        lo, hi = d_score.pitch_range()
        assert lo < hi, "Bad DScore pitch range"
        part_count = d_score.part_count()
        assert part_count == 2, "Bad part count"
        d_part = d_score.combined_d_part()
        part_lo, part_hi = d_part.pitch_range()
        assert part_lo < part_hi, "Bad DPart pitch range"
        assert part_lo == lo, "Bad low pitch found"
        assert part_hi == hi, "Bad high pitch found"
        assert d_part is not None, "Bad DPart retrieval"
        assert d_part.is_monophonic() is False, "Polyphony not detected"
        d_upper = d_score.upper_d_part()
        assert d_upper, "Bad upper DPart"
        assert d_upper.is_monophonic() is True, "Monophonic upper part not detected"
        d_lower = d_score.lower_d_part()
        assert d_lower, "Bad lower DPart"
        # lower_stream = d_lower.stream()
        # lower_stream.show('text')
        assert d_lower.is_monophonic() is True, "Monophonic lower part not detected"

    @staticmethod
    def test_abc_2part_chords():
        p01_path = TestConstant.WTC_CORPUS_DIR + '/prelude01.abc'
        p09_path = TestConstant.WTC_CORPUS_DIR + '/prelude09.abc'
        d_corpus = DCorpus(corpus_path=p01_path)
        score_count = d_corpus.score_count()
        assert score_count == 1, "Bad score count"
        d_corpus.append(corpus_path=p09_path)
        assert d_corpus.score_count() == 2, "Bad append"
        titles = d_corpus.titles()
        assert len(titles) == 2, "Bad title count"
        d_score = d_corpus.d_score_by_index(0)
        assert d_score.title() == 'Prelude 1 (BWV 846)', "Bad title retrieval"
        part_count = d_score.part_count()
        assert part_count == 2, "Bad part count"
        d_part = d_score.combined_d_part()
        assert d_part is not None, "Bad DPart retrieval"
        assert d_part.is_monophonic() is False, "Polyphony not detected"
        d_upper = d_score.upper_d_part()
        assert d_upper, "Bad upper DPart"
        assert d_upper.is_monophonic() is False, "Polyphonic upper part not detected"
        d_lower = d_score.lower_d_part()
        assert d_lower, "Bad lower DPart"
        assert d_lower.is_monophonic() is False, "Polyphonic lower part in Prelude 1 not detected"
        assert d_lower.is_orderly() is False, "Lower part in Prelude 1 is not orderly"
        orderly_stream = d_lower.orderly_note_stream()
        # orderly_stream.show('text')
        orderly_d_part = DPart(music21_stream=orderly_stream)
        assert orderly_d_part.is_orderly() is True, "orderly_note_stream() or is_orderly() is broken"

        # Need to check what happens to tied notes
        d_score = d_corpus.d_score_by_index(1)
        d_upper = d_score.upper_d_part()
        disorderly_stream = d_upper.stream()
        disorderly_d_part = DPart(music21_stream=disorderly_stream)
        assert disorderly_d_part.is_orderly() is False, "orderly_note_stream() or is_orderly() is broken"
        # disorderly_stream.show('text')
        orderly_stream = d_upper.orderly_note_stream()
        orderly_d_part = DPart(music21_stream=orderly_stream)
        assert orderly_d_part.is_orderly() is True, "orderly_note_stream() or is_orderly() is broken"
        # disorderly_stream.show('text')

    @staticmethod
    def test_annotated_corpus():
        da_corpus = DCorpus()
        da_corpus.append_from_db(client_id='695311d7e88d5f79b4945bf45d00cc77', selection_id='21');
        da_score = da_corpus.d_score_by_index(0);
        da_title = da_score.title()
        assert da_title == 'Prelude 2 (BWV 847)', "Bad fetch by index"
        da_score = da_corpus.d_score_by_title(da_title)
        assert da_title == da_score.title(), "Bad fetch by title"
        annotation = da_score.abcd_header()
        abcdf = annotation.abcdf()
        finger_re = re.compile('[12345]+')
        at_re = re.compile('@')
        assert finger_re.search(abcdf), "Bad abcdf"
        assert at_re.search(abcdf), "Bad abcdf"
        upper_abcdf = annotation.upper_abcdf()
        assert finger_re.search(upper_abcdf), "Bad upper abcdf"
        assert not at_re.search(upper_abcdf), "Bad upper abcdf"
        lower_abcdf = annotation.lower_abcdf()
        assert finger_re.search(lower_abcdf), "Bad upper abcdf"
        assert not at_re.search(lower_abcdf), "Bad upper abcdf"
        assert lower_abcdf != upper_abcdf, "Bad split of abcdf"
        abcdf_by_index = da_score.abcdf(index=0)
        abcdf_by_id = da_score.abcdf(identifier=1)
        assert abcdf_by_index == abcdf_by_id, "Bad DSCore::abcdf"
        upper_abcdf_by_index = da_score.upper_abcdf(index=0)
        upper_abcdf_by_id = da_score.upper_abcdf(identifier=1)
        assert upper_abcdf_by_index == upper_abcdf_by_id, "Bad DSCore::upper_abcdf"
        lower_abcdf_by_index = da_score.lower_abcdf(index=0)
        lower_abcdf_by_id = da_score.lower_abcdf(identifier=1)
        assert lower_abcdf_by_index == lower_abcdf_by_id, "Bad DSCore::lower_abcdf"
        # ABCDFAnnotation.ast_for_abcdf(lower_abcdf_by_id)
        # ABCDFAnnotation.ast_for_abcdf(abcdf_by_id)

    @staticmethod
    def test_append_dir():
        d_corpus = DCorpus()
        d_corpus.append_dir(corpus_dir=TestConstant.BERINGER2_ANNOTATED_ARPEGGIO_DIR)
        # d_corpus.append_dir(corpus_dir=TestConstant.BERINGER2_ANNOTATED_SCALE_DIR)
        d_corpus.append_dir(corpus_dir=TestConstant.BERINGER2_ANNOTATED_BROKEN_CHORD_DIR)
        score_count = d_corpus.score_count()
        assert score_count > 48, "Bad score count"
        for i in range(score_count):
            d_score = d_corpus.d_score_by_index(i)
            assert d_score.title() is not None, "Missing title at {0}".format(i)
            assert d_score.is_fully_annotated(indices=[0]), "Missing annotation 1 in {0}".format(d_score.title())


if __name__ == "__main__":
    unittest.main()  # run all tests



__author__ = 'David Randolph'
# Copyright (c) 2014-2018 David A. Randolph.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
import re
from .DAnnotation import DAnnotation


class ABCDHeader:
    COMMENT_RE = r'^%\s*(.*)'
    TITLE_RE = r'^%\s*abcDidactyl v(\d+)'
    FINGERING_RE = r'^%\s*abcD fingering (\d+):\s*(.*)'
    TERMINAL_RE = r'^%\s*abcDidactyl END'
    AUTHORITY_RE = r'^%\s*Authority:\s*([^\(]+)\s*(\((\d+)\))?'
    TRANSCRIBER_RE = r'^%\s*Transcriber:\s*(.*)'
    TRANSCRIPTION_DATE_RE = r'^%\s*Transcription date:\s*((\d\d\d\d\-\d\d\-\d\d)\s*(\d\d:\d\d:\d\d)?)'

    @staticmethod
    def is_abcd(string):
        for line in string.splitlines():
            matt = re.search(ABCDHeader.TITLE_RE, line)
            if matt:
                return True
        return False

    def __init__(self, abcd_str):
        self._annotations = []

        annotation = DAnnotation()
        in_header = False
        for line in abcd_str.splitlines():
            matt = re.search(ABCDHeader.TITLE_RE, line)
            if matt:
                in_header = True
                self._version = matt.group(1)
                continue
            if not in_header:
                continue
            matt = re.search(ABCDHeader.TERMINAL_RE, line)
            if matt:
                break
            matt = re.search(ABCDHeader.FINGERING_RE, line)
            if matt:
                annotation = DAnnotation(abcdf=matt.group(2))
                annotation.abcdf_id(matt.group(1).rstrip())
                self._annotations.append(annotation)
                continue
            matt = re.search(ABCDHeader.AUTHORITY_RE, line)
            if matt:
                annotation.authority(matt.group(1).rstrip())
                if matt.group(2):
                    annotation.authority_year(matt.group(3))
                continue
            matt = re.search(ABCDHeader.TRANSCRIBER_RE, line)
            if matt:
                annotation.transcriber(matt.group(1).rstrip())
                continue
            matt = re.search(ABCDHeader.TRANSCRIPTION_DATE_RE, line)
            if matt:
                annotation.transcription_date(matt.group(1).rstrip())
                continue
            matt = re.search(ABCDHeader.COMMENT_RE, line)
            if matt:
                annotation.add_comment_line(matt.group(1))

    def version(self):
        return self._version

    def annotation_count(self):
        return len(self._annotations)

    def annotations(self):
        return self._annotations

    def annotation_by_id(self, identifier=1):
        for annotation in self._annotations:
            abcdf_id = annotation.abcdf_id()
            if str(abcdf_id) == str(identifier):
                return annotation
        return None

    def annotation(self, index=0, identifier=None):
        if identifier is not None:
            return self.annotation_by_id(identifier)
        if index >= self.annotation_count():
            return None
        return self._annotations[index]

    def abcdf(self, index=0, identifier=None, staff="both"):
        if staff == "both":
            if identifier is not None:
                annotation = self.annotation_by_id(identifier)
                if annotation:
                    return annotation.abcdf()
                else:
                    return None
            if index >= self.annotation_count():
                return None
            return self._annotations[index].abcdf()
        elif staff == "upper":
            return self.upper_abcdf(index=index, identifier=identifier)
        elif staff == "lower":
            return self.lower_abcdf(index=index, identifier=identifier)

        return None

    def upper_abcdf(self, index=0, identifier=None):
        if identifier is not None:
            annotation = self.annotation_by_id(identifier)
            if annotation:
                return annotation.upper_abcdf()
            else:
                return None
        if index >= self.annotation_count():
            return None
        return self._annotations[index].upper_abcdf()

    def lower_abcdf(self, index=0, identifier=None):
        if identifier is not None:
            annotation = self.annotation_by_id(identifier)
            if annotation:
                return annotation.lower_abcdf()
            else:
                return None
        if index >= self.annotation_count():
            return None
        return self._annotations[index].lower_abcdf()



__author__ = 'David Randolph'
# Copyright (c) 2014-2018 David A. Randolph.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
""" Implements and enhances method described in the following paper:

        M. Hart, R. Bosch, and E. Tsai, “Finding Optimal Piano Fingerings,”
            UMAP J., vol. 21, no. 2, pp. 167–177, 2000.
            
    We enhance the method to handle repeated pitches, two staffs,
    and segregated two-hand fingering.
"""

import numpy
import re
import copy
import networkx as nx
import os

from pydactyl.dactyler import Constant
from . import Dactyler as D
from pydactyl.dcorpus.DNote import DNote
BIG_NUM = 999
MAX_INTERVAL_SIZE = 12
BIN_DIR = os.path.abspath(os.path.dirname(__file__))
COST_FILE = Constant.DATA_DIR + '/tables_0.dat'

class Interval:
    def __init__(self, low_color, high_color, low_finger, high_finger, semitone_delta):
        self._l_color = int(low_color)
        self._h_color = int(high_color)
        self._l_finger = int(low_finger)
        self._h_finger = int(high_finger)
        self._s = int(semitone_delta)

    def l_color(self):
        return self._l_color

    def h_color(self):
        return self._h_color

    def l_finger(self):
        return self._l_finger

    def h_finger(self):
        return self._h_finger

    def s(self):
        return self._s

    def __hash__(self):
        val = self._h_finger + 10*self._l_finger + 100*self._h_color + 1000*self._l_color + 10000*self._s
        return val

    def __eq__(self, other):
        if self.l_color() == other.l_color() and self.h_color() == other.h_color() and \
                self.l_finger() == other.l_finger() and self.h_finger() == other.h_finger() and \
                self.s() == other.s():
            return True
        return False

    def __str__(self):
        my_str = "finger: {0}->{1}, color: {2}->{3}, interval: {4}"
        my_str = my_str.format(str(self._l_finger),
                               str(self._h_finger),
                               str(self._l_color),
                               str(self._h_color),
                               str(self._s))
        return my_str

    def __repr__(self):
        return self.__str__()


class Hart(D.Dactyler):
    def _define_costs(self):
        costs = {}
        for l_color in range(2):
            for h_color in range(2):
                for l_finger in range(1, 6):
                    for h_finger in range(1, 6):
                        for s in range(0, self._max_interval_size + 1):
                            # FIXME: What about 0 interval?
                            ic = Interval(l_color, h_color, l_finger, h_finger, s)
                            costs[ic] = BIG_NUM

        interval_size_finalized = False
        max_interval_size = self._max_interval_size
        max_interval_re = re.compile(r"^Max_Interval:\s+(\d+)")
        heading_re = re.compile(r"^([^_]+)_to_([^_]+)")
        cost_re_str = "^(\d)\s+(\d)"
        cost_re = None
        l_color = None
        h_color = None

        f = open(self._cost_path, 'r')
        for line in f:
            line = line.rstrip()
            if not interval_size_finalized:
                matches = re.search(max_interval_re, line)
                if matches:
                    max_interval_size = int(matches.group(1))
                interval_size_finalized = True

                for i in range(max_interval_size + 1):
                    cost_re_str += "\s+(\d+)"
                cost_re_str += "\s*$"
                cost_re = re.compile(cost_re_str)
                continue

            matches = re.search(heading_re, line)
            if matches:
                l_color = Constant.BLACK if matches.group(1) == 'Black' else Constant.WHITE
                h_color = Constant.BLACK if matches.group(2) == 'Black' else Constant.WHITE
                continue

            matches = re.search(cost_re, line)
            if matches:
                l_finger = matches.group(1)
                h_finger = matches.group(2)
                group_num = 3
                for s in range(0, max_interval_size + 1):
                    cost = matches.group(group_num)
                    interval = Interval(low_color=l_color,
                                        high_color=h_color,
                                        low_finger=l_finger,
                                        high_finger=h_finger,
                                        semitone_delta=s)
                    costs[interval] = int(cost)
                    group_num += 1
        f.close()
        return costs

    def __init__(self, segmenter=None, segment_combiner="normal", staff_combiner="naive",
                 cost_path=None, max_interval_size=MAX_INTERVAL_SIZE):
        super().__init__(segmenter=segmenter, segment_combiner=segment_combiner, staff_combiner=staff_combiner)
        self._cost_path = COST_FILE
        if cost_path:
            self._cost_path = cost_path
        self._max_interval_size = max_interval_size
        self._costs = self._define_costs()

    def generate_segment_advice(self, segment, staff, offset=0, cycle=False,
                                handed_first_digit=None, handed_last_digit=None, k=None):
        """
        Generate a set of k ranked fingering suggestions for the given segment. Note that the original
        Hart implementation only returns one best fingering.
        :param segment: The segment to work with, as a music21 score object.
        :param staff: The staff (one of "upper" or "lower") from which the segment was derived.
        :param offset: The zero-based index to begin the returned advice.
        :param cycle: Treat the segment as a repeating pattern and generate advice best suited to
        being repeated. Defaults to False.
        :param handed_first_digit: Constrain the solution to begin with this finger.
        :param handed_last_digit: Constrain the solution to end with this finger.
        :param k: The number of advice segments to return. The actual number returned may be less,
        but will be no more, than this number.
        :return: suggestions, costs, details: Three lists are returned. The first contains suggested fingering
        solutions as abcDF strings. The second contains the respective costs of each suggestion. The third
        contains details on how each cost was determined.
        """
        if k is not None and k != 1:
            raise Exception("Original Hart does not support k best solutions. Try HartK.")
        opt_cost = BIG_NUM

        first_digit = int(handed_first_digit[1:]) if handed_first_digit else None
        last_digit = int(handed_last_digit[1:]) if handed_last_digit else None

        segment_note_count = len(segment)
        note_list = DNote.note_list(segment)
        if len(segment) == 1:
            abcdf = D.Dactyler.one_note_advise(note_list[0], staff=staff,
                                               first_digit=handed_first_digit,
                                               last_digit=handed_last_digit)
            return [abcdf], [0], [0]

        m = segment_note_count - 1
        fs = numpy.zeros([segment_note_count, 6], dtype=int)
        for n in reversed(range(1, segment_note_count)):
            for s in range(1, 6):
                fs[n, s] = BIG_NUM

        fsx = numpy.zeros([segment_note_count, 6, 6], dtype=int)
        num_opt = numpy.zeros([segment_note_count, 6], dtype=int)
        xstar = numpy.zeros([segment_note_count, 6, 5], dtype=int)

        # Stage m
        mth_note = note_list[m]
        mth_color = mth_note.color()
        prior_color = mth_note.prior_color()
        mth_interval = mth_note.semitone_delta()
        penultimate_digit_options = range(1, 6)
        if first_digit and segment_note_count == 2:
            penultimate_digit_options = [first_digit]
        for s in penultimate_digit_options:
            if s == 1:
                self.squawk("Stage {0}: color {1}->{2}, delta {3}".format(m, prior_color, mth_color, mth_interval))
            self.squeak("{0:4d}:".format(s))
            for x in range(1, 6):
                if (staff == "upper" and mth_note.is_ascending()) or (staff == "lower" and not mth_note.is_ascending):
                    interval = Interval(prior_color, mth_color, s, x, mth_interval)
                else:
                    interval = Interval(mth_color, prior_color, x, s, mth_interval)
                cost = self._costs[interval]
                if last_digit:
                    # Last digit in fingering sequence is constrained, so we force all paths to
                    # lead to it by making other paths look less attractive. But we retain preference
                    # for arcs with known reasonable (non-"infinite") costs.
                    if x == last_digit:
                        if cost == BIG_NUM:
                            cost -= 1  # Prefer these arcs over any that lead to the wrong place
                        # Otherwise, prefer plausible arcs (those with non-"infinite" cost) according
                        # to their costs.
                    else:
                        cost = BIG_NUM  # Paths leading to the wrong last digit are "infinitely" expensive.

                fsx[m, s, x] = cost
                self.squeak("{0:4d}  ".format(fsx[m, s, x]))

            for x in range(1, 6):
                if fsx[m, s, x] < fs[m, s]:
                    fs[m, s] = fsx[m, s, x]

            num_opt[m, s] = 0
            for x in range(1, 6):
                if fsx[m, s, x] == fs[m, s]:
                    xstar[m, s, num_opt[m, s]] = x
                    num_opt[m, s] += 1

            for x in range(num_opt[m, s]):
                self.squeak(str(xstar[m, s, x]))
                if x < num_opt[m, s] - 1:
                    self.squeak(",")
                else:
                    self.squawk("")

        # Stages m-1 through 1
        start_index = 1
        for n in reversed(range(1, m)):
            nth_note = note_list[n]
            nth_color = nth_note.color()
            prior_color = nth_note.prior_color()
            nth_interval = nth_note.semitone_delta()
            for s in range(1, 6):
                if s == 1:
                    self.squawk("Stage {0}: color {1}->{2}, delta {3}".format(n, prior_color, nth_color, nth_interval))
                self.squeak("{0:4d}:".format(s))

                for x in range(1, 6):
                    if (staff == "upper" and nth_note.is_ascending()) or \
                            (staff == "lower" and not nth_note.is_ascending):
                        interval = Interval(prior_color, nth_color, s, x, nth_interval)
                    else:
                        interval = Interval(nth_color, prior_color, x, s, nth_interval)
                    cost = self._costs[interval]
                    # First digit in fingering sequence is constrained, so we force all paths to
                    # start from it by making other paths look less attractive. But we retain preference
                    # for arcs with known reasonable (non-"infinite") costs.
                    if first_digit and n == start_index:
                        if s == first_digit:
                            if cost == BIG_NUM:
                                cost -= 1  # Prefer these arcs over any that come from the wrong place
                                # Otherwise, prefer plausible arcs (those with non-"infinite" cost)
                                # according to their costs.
                        else:
                            cost = BIG_NUM  # Paths leading to the wrong last digit are "infinitely" expensive.
                    fsx[n, s, x] = cost + fs[n + 1, x]
                    self.squeak("{0:4d}+{1:d}={2:4d}  ".format(cost, fs[n + 1, x], fsx[n, s, x]))

                for x in range(1, 6):
                    if fsx[n, s, x] < fs[n, s]:
                        fs[n, s] = fsx[n, s, x]

                num_opt[n, s] = 0
                for x in range(1, 6):
                    if fsx[n, s, x] == fs[n, s]:
                        xstar[n, s, num_opt[n, s]] = x
                        num_opt[n, s] += 1

                if num_opt[n, s] == 0:
                    self.squawk("")

                for x in range(num_opt[n, s]):
                    self.squeak(str(xstar[n, s, x]))
                    if x < num_opt[n, s] - 1:
                        self.squeak(",")
                    else:
                        self.squeak("\n")

        fingers = [0]
        for s in range(1, 6):
            fs_cost = fs[1, s]
            if fs_cost < opt_cost:
                opt_cost = fs_cost
                fingers[0] = s
        for n in range(1, m + 1):
            fingers.append(xstar[n, fingers[n - 1], 0])

        self.squawk("The optimal cost is {0}".format(opt_cost))
        self.squawk("Here is an optimal fingering:")
        self.squawk(fingers)

        hand = ">"
        if staff == "lower":
            hand = "<"
        abcdf = hand + "".join(str(f) for f in fingers)
        return [abcdf], [0], [0]


class HartK(Hart):
    def generate_segment_advice(self, segment, staff, offset=0, cycle=False,
                                handed_first_digit=None, handed_last_digit=None, k=None):
        """
        Generate a set of k ranked fingering suggestions for the given segment.
        :param segment: The segment to work with, as a music21 score object.
        :param staff: The staff (one of "upper" or "lower") from which the segment was derived.
        :param offset: The zero-based index to begin the returned advice.
        :param cycle: Treat the segment as a repeating pattern and generate advice best suited to
        being repeated. Defaults to False.
        :param handed_first_digit: Constrain the solution to begin with this finger.
        :param handed_last_digit: Constrain the solution to end with this finger.
        :param k: The number of advice segments to return. The actual number returned may be less,
        but will be no more, than this number.
        :return: suggestions, costs, details: Three lists are returned. The first contains suggested fingering
        solutions as abcDF strings. The second contains the respective costs of each suggestion. The third
        contains details about how each cost was determined. The third is a list of lists containing the
        specific cost of each fingering transition to the corresponding total suggestion cost.
        """
        segment_note_count = len(segment)
        note_list = DNote.note_list(segment)
        if len(segment) == 1:
            abcdf = D.Dactyler.one_note_advise(note_list[0], staff=staff,
                                               first_digit=handed_first_digit,
                                               last_digit=handed_last_digit)
            return [abcdf], [0], [0]

        hand = ">"
        if staff == "lower":
            hand = "<"

        g = nx.DiGraph()
        g.add_node(0, midi=None, handed_digit=None)
        prior_slice_node_ids = list()
        prior_slice_node_ids.append(0)
        last_note_in_segment_index = len(segment) - 1
        node_id = 1
        on_last_prefingered_note = False
        for note_in_segment_index in range(segment_note_count):
            d_note = note_list[note_in_segment_index]
            on_first_prefingered_note = False
            slice_node_ids = list()

            if note_in_segment_index == 0 and handed_first_digit:
                on_first_prefingered_note = True

            if note_in_segment_index == last_note_in_segment_index and handed_last_digit:
                on_last_prefingered_note = True

            for digit in range(1, 6):
                handed_digit = hand + str(digit)
                if on_last_prefingered_note and handed_digit != handed_last_digit:
                    continue
                if on_first_prefingered_note and handed_digit != handed_first_digit:
                    continue

                g.add_node(node_id, midi=d_note.midi(), handed_digit=handed_digit)
                slice_node_ids.append(node_id)
                if 0 in prior_slice_node_ids:
                    g.add_edge(0, node_id, weight=1)
                else:
                    for prior_node_id in prior_slice_node_ids:
                        prior_node = g.nodes[prior_node_id]
                        prior_hd = prior_node["handed_digit"]
                        prior_digit = int(prior_hd[1:])

                        if (staff == "upper" and d_note.is_ascending()) or (staff == "lower" and not d_note.is_ascending):
                            interval = Interval(low_color=d_note.prior_color(),
                                                high_color=d_note.color(),
                                                low_finger=prior_digit,
                                                high_finger=digit,
                                                semitone_delta=d_note.semitone_delta())
                        else:
                            interval = Interval(low_color=d_note.color(),
                                                high_color=d_note.prior_color(),
                                                low_finger=digit,
                                                high_finger=prior_digit,
                                                semitone_delta=d_note.semitone_delta())
                        cost = self._costs[interval]

                        g.add_edge(prior_node_id, node_id, weight=cost)
                node_id += 1
            if len(slice_node_ids) > 0:
                prior_slice_node_ids = copy.copy(slice_node_ids)

        g.add_node(node_id, midi=None, handed_digit=None)
        for prior_node_id in prior_slice_node_ids:
            g.add_edge(prior_node_id, node_id, weight=1)

        return D.Dactyler.generate_standard_graph_advice(g=g, target_id=node_id, k=k)



__author__ = 'David Randolph'
# Copyright (c) 2018 David A. Randolph.
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
import os

TEST_DIR = os.path.abspath(os.path.dirname(__file__))
DATA_DIR = TEST_DIR + "/../data"
CORPORA_DIR = DATA_DIR + "/corpora"
BERINGER2_DIR = CORPORA_DIR + "/beringer2"
BERINGER2_ARPEGGIO_CORPUS = BERINGER2_DIR + "/arpeggios.abc"
BERINGER2_ANNOTATED_ARPEGGIO_DIR = BERINGER2_DIR + "/arpeggios"
BERINGER2_ANNOTATED_BROKEN_CHORD_DIR = BERINGER2_DIR + "/broken_chords"
BERINGER2_ANNOTATED_SCALE_DIR = BERINGER2_DIR + "/scales"

WTC_CORPUS_DIR = CORPORA_DIR + '/wtc'


ONE_NOTE = """% abcDidactyl v5
% abcD fingering 1: 1@
% Authority: Foo Bar (1968)
% Transcriber: David Randolph
% Transcription date: 2016-09-13 17:24:43
% abcD fingering 2: >2@
% Authority: Bar Baz (1066)
% abcDidactyl END
X:1
T:One Note
C:John Doe
M:4/4
K:C
V:1 treble
L:1/4
c4|]"""

TWO_NOTES = """% abcDidactyl v5
% abcD fingering 1: 12@
% Authority: Foo Bar (1968)
% Transcriber: David Randolph
% Transcription date: 2016-09-13 17:24:43
% abcD fingering 2: >23@
% Authority: Bar Baz (1066)
% abcDidactyl END
X:2
T:Dos Notes
C:John Doe
M:4/4
K:C
V:1 treble
L:1/4
b2c2|]"""

ONE_BLACK_NOTE_PER_STAFF = """% abcDidactyl v5
% abcD fingering 1: 12@21
% Authority: Foo Bar (1968)
% Transcriber: David Randolph
% Transcription date: 2016-09-13 17:24:43
% abcD fingering 2: >23@<32
% Authority: Bar Baz (1066)
% abcDidactyl END
X:1
T:One Note Each
C:John Doe
%%score { ( 1 ) | ( 2 ) }
M:4/4
K:C
V:1 treble
V:2 bass octave=-1
V:1 treble
L:1/4
_b4|]
V:2
L:1/4
_E4|]"""

TWO_WHITE_NOTES_PER_STAFF = """% abcDidactyl v5
% abcD fingering 1: 12@21
% Authority: Foo Bar (1968)
% Transcriber: David Randolph
% Transcription date: 2016-09-13 17:24:43
% abcD fingering 2: >23@<32
% Authority: Bar Baz (1066)
% abcDidactyl END
X:1
T:One Note
C:John Doe
%%score { ( 1 ) | ( 2 ) }
M:4/4
K:C
V:1 treble
V:2 bass octave=-1
V:1 treble
L:1/4
C2E2|]
V:2
L:1/4
C,2E,2|]"""

FOUR_NOTES = """% abcDidactyl v5
% abcD fingering 1: 12@21
% Authority: Foo Bar (1968)
% Transcriber: David Randolph
% Transcription date: 2016-09-13 17:24:43
% abcD fingering 2: >5135@<2222
% Authority: Bar Baz (1066)
% abcDidactyl END
X:1
T:Four Note
C:John Doe
%%score { ( 1 ) | ( 2 ) }
M:4/4
K:C
V:1 treble
V:2 bass octave=-1
V:1 treble
L:1/4
ACAe|]
V:2
L:1/4
ACAe|]
"""

A_MAJ_SCALE = """% abcDidactyl v5
% abcD fingering 1: 12312341231234543213214321321&21234123123412321432132143212&12312341231234543213214321321@54321321432132123123412312345&54321321432132123123412312345&32132143213213231231234123121
% Authority:  Beringer and Dunhill (1900)
% Transcriber: David Randolph
% Transcription date: 2016-09-13 17:24:43
% These are complete fingerings, with any gaps filled in.
% abcD fingering 2: >123412312312345431321432143212312312341231231432132143213212341231231234543132143214321@<432143213214321321321432132144321432132143213213214321321421432132143213214321321432132
% Authority: Hart et al. Algorithm
% abcD fingering 3: >123123412312341321421432132112123412312341232132142143213212312341231234132142143213211@<433213214321432341231234521255432132143214323412312345212532132143214321312341231234521
% Authority: Sayegh Algorithm
% abcD fingering 4: >123123412312345432132143213212312312312341232143213213213212312341231234543213214321321@<543214321321321231231234123455432143213213212312312341234532143213213213231231231231234
% Authority: Parncutt Algorithm
% abcDidactyl END
X:7
T:scales_a_major
C:Beringer and Dunhill
%%score { ( 1 ) | ( 2 ) }
M:7/4
K:Amaj
V:1 treble
V:2 bass octave=-1
V:1
L:1/16
A,B,CD EFGA Bcde fgag fedc BAGF EDCB,:|A,4|:
CDEF GABc defg abc'b agfe dcBA GFED:|C4|:
A,B,CD EFGA Bcde fgag fedc BAGF EDCB,:|A,4|]
V:2
L:1/16
A,B,CD EFGA Bcde fgag fedc BAGF EDCB,:|A,4|:
A,B,CD EFGA Bcde fgag fedc BAGF EDCB,:|A,4|:
CDEF GABc [K:clef=treble octave=-1] defg abc'b agfe [K:clef=bass octave=-1] dcBA GFED:|C4|]
"""

A_MAJ_SCALE_SHORT = """% abcDidactyl v5
% abcD fingering 1: 123123412312345@543213214321321
% Authority:  Beringer and Dunhill (1900)
% Transcriber: David Randolph
% Transcription date: 2016-09-13 17:24:43
% These are complete fingerings, with any gaps filled in.
% abcD fingering 2: >123412312312345@<432143213214321
% Authority: Hart et al. Algorithm
% abcD fingering 3: >123123412312341@<433213214321432
% Authority: Sayegh Algorithm
% abcD fingering 4: >123123412312345@<543214321321321
% Authority: Parncutt Algorithm
% abcDidactyl END
X:7
T:scales_a_major
C:Beringer and Dunhill
%%score { ( 1 ) | ( 2 ) }
M:4/4
K:Amaj
V:1 treble
V:2 bass octave=-1
V:1
L:1/16
A,B,CD EFGA Bcde fga2|]
V:2
L:1/16
A,B,CD EFGA Bcde fga2|]
"""

PARNCUTT_HUMAN_MEASURE = dict()
PARNCUTT_HUMAN_MEASURE['A'] = """
X:1
T:A (Op. 821 no. 1)
C:Czerny
M:C
L:1/16
Q:"Allegro"
K:C
V:1 treble
V:2 treble 
V:1
!p!egfg efde cc'bc' abga|]
V:2
[GB]4 z4 [GB]4 z4|]"""

PARNCUTT_HUMAN_MEASURE['B'] = """
X:2
T:B (Op. 821 no. 37)
C:Czerny
M:C
L:1/16
Q:"Allegro"
K:A
V:1 treble
V:2 treble
V:1
!p!ecdf ecdf ecdf ecdf|]
V:2
[A,A]2 z2 [A,A]2 z2 [A,A]2 z2 [A,A]2 z2|]"""

PARNCUTT_HUMAN_MEASURE['C'] = """
X:3
T:C (Op. 821 no. 38)
C:Czerny
M:6/8
L:1/16
Q:"Vivace"
K:A
V:1 treble
V:2 bass
V:1
(3ECE(3AEA(3cAc (3ece(3aea(3c'ac'|]
V:2
A,,2C,2E,2 A,2E,2C,2&A,,12|]"""

PARNCUTT_HUMAN_MEASURE['D'] = """
X:4
T:D (Op. 821 no. 54)
C:Czerny
M:C
L:1/16
Q:"Vivace"
K:C#
V:1 treble
V:2 bass
V:1
!f!CEDF EGFA GBAc Bdce|]
V:2
C,E,D,F, E,G,F,A, G,B,A,C B,DCE|]"""

PARNCUTT_HUMAN_MEASURE['E'] = """
X:5
T:E (Op. 821 no. 62)
C:Czerny
M:C
L:1/16
Q:"Allegro moderato"
K:F
V:1 treble
V:2 bass
V:1
!p!za^g'a' e'f'^c'd' ^gaef ^cd^GA|]
V:2
.D,2.[DF]2.A,2.[DF]2 .D,2.[DF]2.A,2.[DF]2|]"""

PARNCUTT_HUMAN_MEASURE['F'] = """
X:6
T:F (Op. 821 no. 66)
C:Czerny
M:C
L:1/16
Q:"Allegro"
K:Bb
V:1 treble
V:2 bass
V:1
!f!G,B,DB, GDBG dBgd bgd'b|]
V:2
G,,B,,D,B,, G,D,B,G, [K: clef=treble]DB,=GD BGdB|]"""

PARNCUTT_HUMAN_MEASURE['G'] = """
X:7
T:G (Op. 821 no. 96)
C:Czerny
M:3/4
L:1/16
Q:"Allegretto vivace"
K:D
V:1 treble
V:2 bass
V:1
!p!(3f'g'f'(3e'f'e' (3d'e'd'(3c'd'c' (3bc'b(3aba|]
V:2
[B,,B,]2 z2 [F,B,D]2 z2 [F,B,D]2 z2|]"""

PARNCUTT_HUMAN_FRAGMENT = dict()
PARNCUTT_HUMAN_FRAGMENT['A'] = """
X:1
T:A (Op. 821 no. 1)
C:Czerny
M:C
L:1/16
Q:"Allegro"
K:C
V:1 treble
V:2 treble 
V:1
!p!egfg efde z8|]
V:2
[GB]4 z4 z8|]"""

PARNCUTT_HUMAN_FRAGMENT['B'] = """
X:2
T:B (Op. 821 no. 37)
C:Czerny
M:C
L:1/16
Q:"Allegro"
K:A
V:1 treble
V:2 treble
V:1
!p!ecdf z2 z4|]
V:2
[A,A]2 z2 [A,A]2 z2 [A,A]2 z2 [A,A]2 z2|]"""

PARNCUTT_HUMAN_FRAGMENT['C'] = """
X:3
T:C (Op. 821 no. 38)
C:Czerny
M:6/8
L:1/16
Q:"Vivace"
K:A
V:1 treble
V:2 bass
V:1
(3ECE(3AEz(3zzz (3zzz(3zzz(3zzz|]
V:2
A,,2C,2E,2 A,2E,2C,2&A,,12|]"""

PARNCUTT_HUMAN_FRAGMENT['D'] = """
X:4
T:D (Op. 821 no. 54)
C:Czerny
M:C
L:1/16
Q:"Vivace"
K:C#
V:1 treble
V:2 bass
V:1
!f!CEDF EGFz z4 z4|]
V:2
C,E,D,F, E,G,F,A, G,B,A,C B,DCE|]"""

PARNCUTT_HUMAN_FRAGMENT['E'] = """
X:5
T:E (Op. 821 no. 62)
C:Czerny
M:C
L:1/16
Q:"Allegro moderato"
K:F
V:1 treble
V:2 bass
V:1
!p!za^g'a' e'f'^c'd' ^gz3 z4|]
V:2
.D,2.[DF]2.A,2.[DF]2 .D,2.[DF]2.A,2.[DF]2|]"""

PARNCUTT_HUMAN_FRAGMENT['F'] = """
X:6
T:F (Op. 821 no. 66)
C:Czerny
M:C
L:1/16
Q:"Allegro"
K:Bb
V:1 treble
V:2 bass
V:1
!f!G,B,DB, GDz2 z8|]
V:2
G,,B,,D,B,, G,D,B,G, [K: clef=treble]DB,=GD BGdB|]"""

PARNCUTT_HUMAN_FRAGMENT['G'] = """
X:7
T:G (Op. 821 no. 96)
C:Czerny
M:3/4
L:1/16
Q:"Allegretto vivace"
K:D
V:1 treble
V:2 bass
V:1
!p!(3f'g'f'(3e'f'e' (3d'zz(3zzz (3zzz(3zzz|]
V:2
[B,,B,]2 z2 [F,B,D]2 z2 [F,B,D]2 z2|]"""

PARNCUTT_MACHINE = dict()
PARNCUTT_MACHINE['A'] = """% abcDidactyl v5
% abcD fingering 1: ﻿24342313@xxxx
% Authority: Parncutt 1
% Transcriber: David Randolph
% Transcription date: 2018-06-19 17:24:43
% abcDidactyl END
X:1
T:A (Op. 821 no. 1)
C:Czerny
M:C
L:1/16
Q:"Allegro"
K:C
V:1 treble
V:2 treble 
V:1
!p!egfg efde cc'bc' abga|]
V:2
[GB]4 z4 [GB]4 z4|]"""



#!/usr/bin/env python3
import re
import unittest
from pydactyl.dactyler.Parncutt import Parncutt
from pydactyl.dcorpus.DCorpus import DCorpus
from pydactyl.dcorpus.ManualDSegmenter import ManualDSegmenter
import TestConstant

last_digit = {
    'A': 3,
    # B is cyclic pattern
    'C': 1,
    'D': 2,
    'E': 2,
    'F': 1,
    'G': 1
}

sma_lar_sums = {
    'A': {
        '>24342313': 1, '>13231213': 1, '>14342313': 3, '>35453423': 0, '>13231323': 4,
        '>13231423': 4, '>13242313': 3, '>12142313': 4, '>13232313': 3, '>24342323': 2
    },
    'B': {
        '>3124': 0, '>3214': 5, '>4235': 0, '>4215': 5, '>1235': 9,
        '>2124': 2, '>3125': 4, '>3215': 9, '>3235': 4, '>4125': 4
    },
    'C': {
        '>21241': 2, '>21251': 2, '>31241': 2, '>31251': 2, '>32351': 8,
        '>42351': 6, '>31351': 4, '>42151': 8, '>21351': 4, '>32151': 10
    },
    'D': {
        '>1323132': 4, '>1324132': 4, '>2412132': 3, '>1212132': 0, '>1213242': 2,
        '>1323142': 8, '>1323242': 4, '>2413242': 5, '>2435132': 6,'>1213232': 2
    },
    'E': {
        '>14523152': 8, '>14523142': 10, '>14512152': 10, '>14513152': 12, '>14515342': 14,
        '>14515352': 14, '>14524152': 13, '>13412152': 10, '>13523152': 14, '>14515452': 14
    },
    'F': {
        '>124251': 8, '>124151': 3, '>123151': 6, '>124141': 1, '>123141': 4,
        '>124131': 3, '>135141': 3, '>135251': 10, '>125251': 12, '>135131': 5
    },
    'G': {
        '>3432321': 0, '>2321321': 1, '>3431321': 2, '>2432321': 4, '>4542321': 2,
        '>2132321': 6, '>2321421': 5, '>2321431': 4, '>3132321': 8, '>3432421': 4
    }
}

subcosts = {
    'A': {
        '>24342313': {'str': 0, 'wea': 2, '345': 0, '3t4': 1, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>13231213': {'str': 0, 'wea': 0, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>14342313': {'str': 0, 'wea': 2, '345': 0, '3t4': 1, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>35453423': {'str': 0, 'wea': 4, '345': 3, '3t4': 1, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>13231323': {'str': 0, 'wea': 0, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>13231423': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>13242313': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>12142313': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>13232313': {'str': 0, 'wea': 0, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>24342323': {'str': 0, 'wea': 2, '345': 0, '3t4': 1, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0}
    },
    # For Piece B, weak finger (wea) costs in published Parncutt paper is one less than
    # our values when the first finger is a weak finger.
    # Bad wea cost for >4235 in B: 3 should be 2
    # Bad wea cost for >4215 in B: 3 should be 2
    # Bad wea cost for >4125 in B: 3 should be 2
    #
    # But all other costs are the same as ours. ???? FIXME ???? Is Parncutt including
    # the costs for the transition from the fourth note to the first or not?
    #
    # 345
    'B': {
        '>3124': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 1, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>3214': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 1, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>4235': {'str': 0, 'wea': 3, '345': 1, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 4, 'pa1': 0},
        '>4215': {'str': 0, 'wea': 3, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 4, 'pa1': 0},
        '>1235': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 4, 'pa1': 0},
        '>2124': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>3125': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 4, 'pa1': 0},
        '>3215': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 4, 'pa1': 0},
        '>3235': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 4, 'pa1': 0},
        '>4125': {'str': 0, 'wea': 3, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 4, 'pa1': 0},
        '>1234': {'str': 4, 'wea': 1, '345': 0, '3t4': 1, 'bl4': 1, 'bl1': 0, 'bl5': 0, 'pa1': 0},
    },
    'C': {
        '>21241': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>21251': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>31241': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>31251': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>32351': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>42351': {'str': 0, 'wea': 2, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>31351': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>42151': {'str': 0, 'wea': 2, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>21351': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>32151': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>21231': {'str': 4, 'wea': 0, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
    },
    'D': {
        '>1323132': {'str': 0, 'wea': 0, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 3, 'bl5': 0, 'pa1': 0},
        '>1324132': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 3, 'bl5': 0, 'pa1': 0},
        '>2412132': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 3, 'bl5': 0, 'pa1': 0},
        '>1212132': {'str': 0, 'wea': 0, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 6, 'bl5': 0, 'pa1': 0},
        '>1213242': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 6, 'bl5': 0, 'pa1': 0},
        '>1323142': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 3, 'bl5': 0, 'pa1': 0},
        '>1323242': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 3, 'bl5': 0, 'pa1': 0},
        '>2413242': {'str': 0, 'wea': 2, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 3, 'bl5': 0, 'pa1': 0},
        '>2435132': {'str': 0, 'wea': 2, '345': 1, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 2, 'pa1': 0},
        '>1213232': {'str': 0, 'wea': 0, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 6, 'bl5': 0, 'pa1': 0},
    },
    'E': {
        '>14523152': {'str': 0, 'wea': 3, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>14523142': {'str': 2, 'wea': 3, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>14512152': {'str': 0, 'wea': 3, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>14513152': {'str': 0, 'wea': 3, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>14515342': {'str': 2, 'wea': 4, '345': 1, '3t4': 1, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>14515352': {'str': 0, 'wea': 4, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>14524152': {'str': 0, 'wea': 4, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>13412152': {'str': 2, 'wea': 2, '345': 0, '3t4': 1, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>13523152': {'str': 2, 'wea': 2, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>14515452': {'str': 2, 'wea': 5, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>13415342': {'str': 4, 'wea': 3, '345': 1, '3t4': 2, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
    },
    'F': {
        '>124251': {'str': 2, 'wea': 2, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>124151': {'str': 0, 'wea': 2, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>123151': {'str': 2, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>124141': {'str': 0, 'wea': 2, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>123141': {'str': 2, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>124131': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>135141': {'str': 0, 'wea': 2, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>135251': {'str': 2, 'wea': 2, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>125251': {'str': 2, 'wea': 2, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>135131': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 5, 'bl5': 0, 'pa1': 0},
        '>123251': {'str': 6, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
    },
    'G': {
        '>3432321': {'str': 0, 'wea': 1, '345': 0, '3t4': 1, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>2321321': {'str': 0, 'wea': 0, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>3431321': {'str': 0, 'wea': 1, '345': 0, '3t4': 1, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>2432321': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>4542321': {'str': 0, 'wea': 3, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>2132321': {'str': 0, 'wea': 0, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>2321421': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>2321431': {'str': 0, 'wea': 1, '345': 0, '3t4': 0, 'bl4': 1, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>3132321': {'str': 0, 'wea': 0, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>3432421': {'str': 0, 'wea': 2, '345': 0, '3t4': 1, 'bl4': 0, 'bl1': 0, 'bl5': 0, 'pa1': 0},
        '>1321321': {'str': 0, 'wea': 0, '345': 0, '3t4': 0, 'bl4': 0, 'bl1': 3, 'bl5': 0, 'pa1': 0},
        '>3432431': {'str': 0, 'wea': 2, '345': 0, '3t4': 1, 'bl4': 1, 'bl1': 0, 'bl5': 0, 'pa1': 0},
    }
}

subcost_names = ('str', 'sma', 'lar', 'pcc', 'pcs', 'wea', '345', '3t4', 'bl4', 'bl1', 'bl5', 'pa1')

solutions = {
    'A': {
        6: {'>24342313': True},
        7: {'>13231213': True},
        8: {'>14342313': True, '>35453423': True},
        9: {'>13231323': True, '>13231423': True, '>13242313': True},
        10: {'>12142313': True, '>13232313': True, '>24342323': True}
    },
    'B': {
        7: {'>3124': True, '>3214': True, '>4235': True},
        11: {'>4215': True},
        14: {'>1235': True, '>2124': True, '>3125': True, '>3215': True},
        15: {'>3235': True, '>4125': True}
    },
    'C': {
        9: {'>21241': True, '>21251': True},
        10: {'>31241': True, '>31251': True, '>32351': True, '>42351': True},
        11: {'>31351': True, '>42151': True},
        12: {'>21351': True, '>32151': True}
    },
    'D': {
        12: {'>1323132': True, '>1324132': True},
        13: {'>2412132': True},
        14: {'>1212132': True, '>1213242': True, '>1323142': True},
        15: {'>1323242': True, '>2413242': True, '>2435132': True},
        16: {'>1213232': True}
    },
    'E': {
        24: {'>14523152': True},
        28: {'>14523142': True},
        31: {'>14512152': True},
        32: {'>14513152': True, '>14515342': True, '>14515352': True, '>14524152': True},
        33: {'>13412152': True, '>13523152': True, '>14515452': True}
    },
    'F': {
        17: {'>124251': True},
        19: {'>124151': True},
        20: {'>123151': True, '>124141': True},
        22: {'>123141': True, '>124131': True},
        23: {'>135141': True, '>135251': True},
        24: {'>125251': True, '>135131': True}
    },
    'G': {
        2: {'>3432321': True},
        3: {'>2321321': True},
        5: {'>3431321': True},
        6: {'>2432321': True, '>4542321': True},
        7: {'>2132321': True},
        8: {'>2321421': True, '>2321431': True, '>3132321': True, '>3432421': True}
    }
}


class ParncuttTest(unittest.TestCase):
    def test_four_note_example(self):
        parncutt = Parncutt(segmenter=ManualDSegmenter(), segment_combiner="cost")
        # parncutt.segment_combiner(method="cost")
        d_corpus = DCorpus(corpus_str=TestConstant.FOUR_NOTES)
        parncutt.load_corpus(d_corpus=d_corpus)
        suggestions, costs, details = parncutt.generate_advice(staff="upper", k=2)
        self.assertEqual(len(suggestions), 2, "No loops in that dog in top ten")
        parncutt.report_on_advice(suggestions, costs, details)

    # @staticmethod
    # def test_cycles():
    #     parncutt = Parncutt()
    #     parncutt.segment_combination_method(method="cost")
    #     d_corpus = DCorpus(corpus_str=TestConstant.FOUR_NOTES)
    #     parncutt.load_corpus(d_corpus=d_corpus)
    #     suggestions, costs, details = parncutt.generate_advice(staff="upper", cycle=4, k=2)
    #     assert len(suggestions) == 2, "No loops in that dog in top ten"
    #     parncutt.report_on_advice(suggestions, costs, details)
    #     d_corpus = DCorpus(corpus_str=TestConstant.PARNCUTT_HUMAN_FRAGMENT['B'])
    #     parncutt.load_corpus(d_corpus=d_corpus)
    #     suggestions, costs, details = parncutt.generate_advice(staff="upper", cycle=4, k=16)
    #     assert len(suggestions) == 16, "There should be 16 cyclic fingerings!"
    #     parncutt.report_on_advice(suggestions, costs, details)
    #
    # @staticmethod
    # def test_fingering_counts():
    #     parncutt = Parncutt(pruning_method="none")
    #     parncutt.segment_combination_method(method="cost")
    #     d_corpus = DCorpus(corpus_str=TestConstant.FOUR_NOTES)
    #     parncutt.load_corpus(d_corpus=d_corpus)
    #     suggestions, costs, details = parncutt.generate_advice(staff="upper", k=2)
    #     assert parncutt.pruned_count() == 320, "Bad none pruning on open-ended problem"
    #     suggestions, costs, details = parncutt.generate_advice(staff="upper", last_digit=5, k=10)
    #     print(suggestions)
    #     print(parncutt.pruned_count())
    #
    # @staticmethod
    # def test_sma_lar():
    #     parncutt = Parncutt()
    #     parncutt.segment_combination_method(method="cost")
    #
    #     for id in sma_lar_sums:
    #         d_corpus = DCorpus(corpus_str=TestConstant.PARNCUTT_HUMAN_FRAGMENT[id])
    #         parncutt.load_corpus(d_corpus=d_corpus)
    #         if id == 'B':
    #             suggestions, costs, details = parncutt.generate_advice(staff="upper", cycle=4, k=20)
    #         else:
    #             suggestions, costs, details = parncutt.generate_advice(staff="upper", last_digit=last_digit[id], k=20)
    #         details_for_sugg = dict()
    #         for i in range(len(details)):
    #             details_for_sugg[suggestions[i]] = details[i][0]  # 0 index because we only have one segment
    #
    #         # parncutt.report_on_advice(suggestions, costs, details)
    #         for gold_sugg in sma_lar_sums[id]:
    #             assert gold_sugg in details_for_sugg,\
    #                 "Missing suggestion {0} in {1}".format(gold_sugg, id)
    #             sma = details_for_sugg[gold_sugg]['sma']
    #             lar = details_for_sugg[gold_sugg]['lar']
    #             assert sma + lar == sma_lar_sums[id][gold_sugg],\
    #                 "Bad sma + lar total for {0} in {1}".format(gold_sugg, id)
    #
    # @staticmethod
    # def test_bl1():
    #     parncutt = Parncutt()
    #     parncutt.segment_combination_method(method="cost")
    #     midi_1 = None
    #     handed_digit_1 = '-'
    #     midi_2 = 61
    #     handed_digit_2 = '>1'
    #     midi_3 = 65
    #     handed_digit_3 = '>3'
    #
    #     cost, costs = parncutt.trigram_node_cost(midi_1, handed_digit_1, midi_2, handed_digit_2, midi_3, handed_digit_3)
    #     assert costs['bl1'] == 3, "Bad bl1 cost"
    #
    # @staticmethod
    # def test_good_rules():
    #     parncutt = Parncutt()
    #     parncutt.segment_combination_method(method="cost")
    #
    #     for id in subcosts:
    #         d_corpus = DCorpus(corpus_str=TestConstant.PARNCUTT_HUMAN_FRAGMENT[id])
    #         parncutt.load_corpus(d_corpus=d_corpus)
    #         if id == 'B':
    #             suggestions, costs, details = parncutt.generate_advice(staff="upper", cycle=4, k=20)
    #         else:
    #             suggestions, costs, details = parncutt.generate_advice(staff="upper", last_digit=last_digit[id], k=30)
    #         details_for_sugg = dict()
    #         for i in range(len(details)):
    #             details_for_sugg[suggestions[i]] = details[i][0]  # 0 index because we only have one segment
    #
    #         parncutt.report_on_advice(suggestions, costs, details)
    #         for gold_sugg in subcosts[id]:
    #             assert gold_sugg in details_for_sugg, \
    #                 "Missing suggestion {0} in {1}".format(gold_sugg, id)
    #             for rule in subcosts[id][gold_sugg]:
    #                 # if rule == 'bl1':
    #                     # continue
    #                 gold_cost = subcosts[id][gold_sugg][rule]
    #                 cost = details_for_sugg[gold_sugg][rule]
    #                 assert cost == gold_cost, \
    #                     "Bad {0} cost for {1} in {2}: {3} should be {4}".format(rule, gold_sugg, id, cost, gold_cost)

    #

    # @staticmethod
    # def test_corrected_parncutt():
    #     parncutt = Parncutt(pruning_method="none")
    #     parncutt.segment_combination_method(method="cost")
    #
    #     for id in subcosts:
    #         if id != 'B':
    #             continue
    #         d_corpus = DCorpus(corpus_str=TestConstant.PARNCUTT_HUMAN_FRAGMENT[id])
    #         parncutt.load_corpus(d_corpus=d_corpus)
    #         if id == 'B':
    #             suggestions, costs, details = parncutt.generate_advice(staff="upper", cycle=4, k=500)
    #         else:
    #             suggestions, costs, details = parncutt.generate_advice(staff="upper", last_digit=last_digit[id], k=500)
    #         playable_count = parncutt.pruned_count()
    #         # parncutt.report_on_advice(suggestions, costs, details)
    #
    #         for i in range(len(details)):
    #             sugg = suggestions[i]
    #             cost = costs[i]
    #             subcosts_for_sugg = details[i][0]  # 0 index because we only have one segment
    #             record_str = "{0}&{1}&".format(i+1, sugg[1:])
    #             for subcost_name in subcost_names:
    #                 record_str += str(subcosts_for_sugg[subcost_name]) + '&'
    #             record_str += str(int(cost)) + "\\\\"
    #             print(record_str)
    #         print("Piece {0}, playable fingerings: {1}".format(id, playable_count))

        # for id in solutions:
        #     d_corpus = DCorpus(corpus_str=TestConstant.PARNCUTT_HUMAN_FRAGMENT[id])
        #     parncutt.load_corpus(d_corpus=d_corpus)
        #     suggestions, costs, details = parncutt.generate_advice(staff="upper", last_digit=last_digit[id], k=20)
        #     parncutt.report_on_advice(suggestions, costs, details)
        #     for i in range(10):
        #         assert costs[i] in solutions[id],\
        #             "Missing cost {0} in {1}".format(costs[i], id)
        #         assert suggestions[i] in solutions[id][int(costs[i])],\
        #             "Missing {0} cost suggestion {1} in {2}".format(costs[i], suggestions[i], id)


    # @staticmethod
    # def test_parncutt_edges():
    #     parncutt = Parncutt()
    #     d_corpus = DCorpus(corpus_str=TestConstant.ONE_NOTE)
    #     parncutt.load_corpus(d_corpus=d_corpus)
    #     upper_rh_advice = parncutt.advise(staff="upper")
    #     right_re = re.compile('^>\d$')
    #     assert right_re.match(upper_rh_advice), "Bad one-note, right-hand, upper-staff advice"
    #     # both_advice = parncutt.advise(staff="both")
    #     # both_re = re.compile('^>\d@$')
    #     # assert both_re.match(both_advice), "Bad one-note, segregated, both-staff advice"
    #
    #     parncutt = Parncutt()
    #     d_corpus = DCorpus(corpus_str=TestConstant.ONE_BLACK_NOTE_PER_STAFF)
    #     parncutt.load_corpus(d_corpus=d_corpus)
    #     upper_advice = parncutt.advise(staff="upper")
    #
    #     right_re = re.compile('^>2$')
    #     assert right_re.match(upper_advice), "Bad black-note, upper-staff advice"
    #     lower_advice = parncutt.advise(staff="lower")
    #     left_re = re.compile('^<2$')
    #     assert left_re.match(lower_advice), "Bad black-note, upper-staff advice"
    #     both_advice = parncutt.advise(staff="both")
    #     both_re = re.compile('^>2@<2$')
    #     assert both_re.match(both_advice), "Bad black-note, both-staff advice"
    #     lower_advice = parncutt.advise(staff="lower", first_digit=3)
    #     lower_re = re.compile('^<3$')
    #     assert lower_re.match(lower_advice), "Bad preset black-note, both-staff advice"
    #
    #     parncutt = Parncutt()
    #     d_corpus = DCorpus(corpus_str=TestConstant.TWO_WHITE_NOTES_PER_STAFF)
    #     parncutt.load_corpus(d_corpus=d_corpus)
    #     upper_advice = parncutt.advise(staff="upper")
    #     right_re = re.compile('^>\d\d$')
    #     assert right_re.match(upper_advice), "Bad two white-note, upper-staff advice"
    #     upper_advice = parncutt.advise(staff="upper", first_digit=2, last_digit=4)
    #     right_re = re.compile('^>24$')
    #     assert right_re.match(upper_advice), "Bad preset two white-note, upper-staff advice"
    #
    # @staticmethod
    # def test_distance_metrics():
    #     parncutt = Parncutt()
    #     d_corpus = DCorpus(corpus_str=TestConstant.A_MAJ_SCALE)
    #     parncutt.load_corpus(d_corpus=d_corpus)
    #     complete_rh_advice = parncutt.advise(staff="upper")
    #     complete_rh_advice_len = len(complete_rh_advice)
    #     right_re = re.compile('^>\d+$')
    #     assert right_re.match(complete_rh_advice), "Bad right-hand, upper-staff advice"
    #     rh_advice = parncutt.advise(staff="upper", offset=3, first_digit=4)
    #     short_advice_len = len(rh_advice)
    #     assert complete_rh_advice_len - 3 == short_advice_len, "Bad offset for advise() call"
    #     ff_re = re.compile('^>4\d+$')
    #     assert ff_re.match(rh_advice), "Bad first finger constraint"
    #
    #     rh_advice = parncutt.advise(staff="upper", offset=10, first_digit=1, last_digit=3)
    #     short_advice_len = len(rh_advice)
    #     assert complete_rh_advice_len - 10 == short_advice_len, "Bad offset for advise() call"
    #     ff_re = re.compile('^>1\d+3$')
    #     assert ff_re.match(rh_advice), "Bad first and last finger constraints"
    #
    #     lh_advice = parncutt.advise(staff="lower")
    #     left_re = re.compile('^<\d+$')
    #     assert left_re.match(lh_advice), "Bad left-hand, lower-staff advice"
    #     combo_advice = parncutt.advise(staff="both")
    #     clean_combo_advice = re.sub('[><&]', '',  combo_advice)
    #     d_score = d_corpus.d_score_by_index(index=0)
    #     gold_fingering = d_score.abcdf(index=0)
    #     clean_gold_fingering = re.sub('[><&]', '',  gold_fingering)
    #
    #     combo_re = re.compile('^>\d+@<\d+$')
    #     assert combo_re.match(combo_advice), "Bad combined advice"
    #     hamming_evaluations = parncutt.evaluate_strike_distance(method="hamming", staff="both")
    #     assert hamming_evaluations[0] > 0, "Undetected Hamming costs"
    #     assert hamming_evaluations[3] == 0, "Bad fish in Hamming barrel"
    #
    #     natural_evaluations = parncutt.evaluate_strike_distance(method="natural", staff="both")
    #     assert natural_evaluations[0] > 0, "Undetected natural costs"
    #     assert natural_evaluations[3] == 0, "Bad fish in natural barrel"
    #
    #     pivot_evaluations = parncutt.evaluate_strike_distance(method="pivot", staff="both")
    #     assert pivot_evaluations[0] > 0, "Undetected pivot costs"
    #     assert pivot_evaluations[3] == 0, "Bad fish in pivot barrel"
    #
    #
    # @staticmethod
    # def test_reentry():
    #     parncutt = Parncutt()
    #
    #     # We cannot use the longer example A_MAJ_SCALE because the gold standard fingering
    #     # requires hand repositionings not allowed by the Parncutt model. This reinforces the need
    #     # for segmentation and also (maybe) the need for a more inclusive option for Parncutt where
    #     # all paths are possible but some are just very expensive, as we have in Sayegh.
    #     d_corpus = DCorpus(corpus_str=TestConstant.A_MAJ_SCALE_SHORT)
    #     parncutt.load_corpus(d_corpus=d_corpus)
    #
    #     reentry_hamming_evals = parncutt.evaluate_strike_reentry(method="hamming", staff="upper", gold_indices=[2, 3])
    #     # Note we are not picking Beringer for the real gold standard because Beringer and Parncutt agree
    #     # on the fingering for this scale.
    #     # for rhe in reentry_hamming_evals:
    #         # print("RHE:{0}".format(rhe))
    #     assert reentry_hamming_evals[0] > 0, "Undetected upper Hamming reentry costs"
    #     assert reentry_hamming_evals[1] == 0, "Bad fish in upper-staff Hamming reentry barrel"
    #
    #     reentry_hamming_evals = parncutt.evaluate_strike_reentry(method="hamming", staff="both", gold_indices=[2, 3])
    #     # for rhe in reentry_hamming_evals:
    #         # print("RHE:{0}".format(rhe))
    #     assert reentry_hamming_evals[0] > 0, "Undetected both-staff Hamming reentry costs"
    #     assert reentry_hamming_evals[1] == 0, "Bad fish in both-staff Hamming reentry barrel"
    #     hamming_score = reentry_hamming_evals[0]
    #
    #     reentry_natural_evals = parncutt.evaluate_strike_reentry(method="natural", staff="both", gold_indices=[2, 3])
    #     # for rne in reentry_natural_evals:
    #         # print("RNE:{0}".format(rne))
    #     assert reentry_natural_evals[0] > 0, "Undetected natural reentry costs"
    #     assert reentry_natural_evals[1] == 0, "Bad fish in natural reentry barrel"
    #     natural_score = reentry_natural_evals[0]
    #     assert natural_score > hamming_score, "Reentry: Natural <= Hamming"
    #
    #     reentry_pivot_evals = parncutt.evaluate_strike_reentry(method="pivot", staff="both", gold_indices=[2, 3])
    #     # for rpe in reentry_pivot_evals:
    #         # print("RPE:{0}".format(rpe))
    #     assert reentry_pivot_evals[0] > 0, "Undetected pivot reentry costs"
    #     assert reentry_pivot_evals[1] == 0, "Bad fish in pivot reentry barrel"
    #     pivot_score = reentry_pivot_evals[0]
    #     assert natural_score < pivot_score, "Reentry: Natural >= Pivot"
    #
    # @staticmethod
    # def test_pivot_alignment():
    #     parncutt = Parncutt()
    #     d_corpus = DCorpus(corpus_str=TestConstant.A_MAJ_SCALE)
    #     parncutt.load_corpus(d_corpus=d_corpus)
    #
    #     evaluations = parncutt.evaluate_pivot_alignment(staff="both")
    #     # for he in hamming_evaluations:
    #     # print(he)
    #     assert evaluations[0] > 0, "Undetected pivot alignment costs"
    #     assert evaluations[3] == 0, "Bad fish in pivot alignment barrel"


if __name__ == "__main__":
    unittest.main()  # run all tests
